# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib, os, dill, time, inspect, json as _json, base64; k1 = k1lib
from typing import List
import k1lib.cli as cli; from k1lib.cli import *
from collections import defaultdict
pygments = k1.dep("pygments")
try: import PIL.Image, PIL; hasPIL = True
except: PIL = k1.dep("PIL"); hasPIL = False
__all__ = ["tag_serve",
           "FromNotebook", "FromPythonFile", "BuildPythonFile", "BuildDashFile", "StartServer", "GenerateHtml", "commonCbs", "serve",
           "text", "slider", "html", "json", "date", "serialized", "apiKey", "analyze", "webToPy", "pyToWeb"]
basePath = os.path.dirname(inspect.getabsfile(k1lib)) + os.sep + "serve" + os.sep
def pretty_py(code_string):                                                      # pretty_py
    s = pygments.highlight(code_string, pygments.lexers.PythonLexer(), pygments.formatters.HtmlFormatter()).replace('<pre', '<pre class="pre"') # pretty_py
    css = pygments.formatters.HtmlFormatter().get_style_defs('.highlight').replace(".highlight { background: #f8f8f8; }", "") # pretty_py
    return f'<style>{css}</style>{s}'                                            # pretty_py
def pretty_js(code_string):                                                      # pretty_js
    from pygments.lexers.javascript import JavascriptLexer                       # pretty_js
    s = pygments.highlight(code_string, JavascriptLexer(), pygments.formatters.HtmlFormatter()).replace('<pre', '<pre class="pre"') # pretty_js
    css = pygments.formatters.HtmlFormatter().get_style_defs('.highlight').replace(".highlight { background: #f8f8f8; }", "") # pretty_js
    return f'<style>{css}</style>{s}'                                            # pretty_js
def execPy(s:str, node:str, condaEnv:str=None, condaPath:str="~/miniconda3") -> "list[str]": # execPy
    fn = s | cli.file(); None | cli.cmd(f"scp {fn} {node}:{fn}") | cli.deref()   # execPy
    return None | cli.cmd(f"ssh {node} 'source {condaPath}/bin/activate {condaEnv} && python {fn}'") | cli.filt("x") | cli.deref() # execPy
def tag_serve(node:str=None, condaEnv:str=None, condaPath:str="~/miniconda3"):   # tag_serve
    """Tag that marks the cell that will be extracted to an independent
file and executed. Example::

    # serve(node="user@hostname", condaEnv="torch")

If a remote node is specified, internally, this will run commands on that node using
ssh, so make sure it's reachable via "ssh user@hostname" with default ssh identity,
or else it wouldn't work.

If .condaEnv is specified, then will activate conda before executing the script.
The activation command will be "{condaPath}/bin/activate {condaEnv}"

:param node: what node should the script be executed?
:param condaEnv: if specified, will activate that conda env before executing"""  # tag_serve
    return {"node": node, "condaEnv": condaEnv}                                  # tag_serve
class FromNotebook(k1.Callback):                                                 # FromNotebook
    def __init__(self, fileName, tagName:str="serve", allTags:"list[str]"=("test", "notest", "donttest", "thumbnail", "export", "serve", "noserve", "dash")): # FromNotebook
        """Grabs source code from a Jupyter notebook. Will grab cells with the comment
like ``# serve`` in the first line.

See :meth:`tag_serve` to see more about its options

:param fileName: notebook path
:param tagName: which tag to extract out?
:param allTags: all possible custom tags that you use. It might complain if there's a tag in your
    notebook that it doesn't know about, so add all of your commonly used tags here""" # FromNotebook
        super().__init__(); self.fileName = fileName; self.tagName = tagName; self.allTags = allTags # FromNotebook
    def fetchSource(self):                                                       # FromNotebook
        a = cli.nb.cells(self.fileName) | cli.filt(cli.op()["cell_type"] == "code") | cli.aS(list); self.l["sourceType"] = "notebook" # FromNotebook
        self.l["sourceCode"] = a | cli.nb.pretty(whitelist=[self.tagName]) | (cli.op()["source"] | ~cli.head(1)).all() | cli.joinStreams() | cli.deref() # FromNotebook
        self.l["tags"] = a | cli.op()["source"].all() | cli.filt("x") | cli.item().all() | cli.filt(cli.op().startswith("#")) | cli.deref() # FromNotebook
        # figures out build vars here, like node & condaEnv                      # FromNotebook
        self.l["node"] = None; self.l["condaEnv"] = None; self.l["condaPath"] = "~/miniconda3" # FromNotebook
        def serve(node:str=None, condaEnv:str=None, condaPath:str="~/miniconda3"): self.l["node"] = self.l["node"] or node; self.l["condaEnv"] = self.l["condaEnv"] or condaEnv; self.l["condaPath"] = self.l["condaPath"] or condaPath # FromNotebook
        for tag in self.l["tags"]: nb.executeTags(tag, defaultdict(lambda: 0, {**{x:0 for x in self.allTags}, "serve": serve})) # FromNotebook
class FromPythonFile(k1.Callback):                                               # FromPythonFile
    def __init__(self, fileName):                                                # FromPythonFile
        """Grabs source code from a python file."""                              # FromPythonFile
        super().__init__(); self.fileName = fileName                             # FromPythonFile
    def fetchSource(self): self.l["sourceCode"] = cli.cat(self.fileName) | cli.deref(); self.l["sourceType"] = "file" # FromPythonFile
class BuildPythonFile(k1.Callback):                                              # BuildPythonFile
    def __init__(self, port=None):                                               # BuildPythonFile
        """Builds the output Python file, ready to be served on localhost.

:param port: which port to run on localhost. If not given, then a port will
    be picked at random, and will be available at ``cbs.l['port']``"""           # BuildPythonFile
        super().__init__(); self.port = port; self.suffix = "suffix.py"          # BuildPythonFile
    def buildPythonFile(self):                                                   # BuildPythonFile
        self.l["pythonFile"] = ["from k1lib.imports import *", *self.l["sourceCode"]] | cli.file(); port = self.port # BuildPythonFile
        self.l["metaFile"] = metaFile = "" | cli.file(); os.remove(metaFile) # grabs temp meta file for communication, on localhost, not remote # BuildPythonFile
        if self.l["node"] is None: # grabs random free port if one is not available # BuildPythonFile
            if port is None: import socket; sock = socket.socket(); sock.bind(('', 0)); port = sock.getsockname()[1]; sock.close() # BuildPythonFile
        else:                                                                    # BuildPythonFile
            kw = {"node": self.l["node"], "condaEnv": self.l["condaEnv"], "condaPath": self.l["condaPath"]} # BuildPythonFile
            if port is None: port = execPy('import socket; sock = socket.socket(); sock.bind(("", 0)); port = sock.getsockname()[1]; sock.close(); print(port)', **kw)[0] # BuildPythonFile
        # actually has enough info to build the final file                       # BuildPythonFile
        self.l["port"] = port; node = self.l["node"]; (cli.cat(f"{basePath}{self.suffix}") | cli.op().replace("SOCKET_PORT", f"{port}").replace("META_FILE", metaFile).all()) >> cli.file(self.l["pythonFile"]) # BuildPythonFile
        if node: fn = self.l["pythonFile"]; None | cli.cmd(f"scp {fn} {node}:{fn}") | cli.deref() # BuildPythonFile
class BuildDashFile(BuildPythonFile):                                            # BuildDashFile
    def __init__(self):                                                          # BuildDashFile
        """Builds the output Python file for a Dash app, ready to be served on localhost""" # BuildDashFile
        super().__init__()                                                       # BuildDashFile
        self.suffix = "suffix-dash.py"                                           # BuildDashFile
class StartServer(k1.Callback):                                                  # StartServer
    def __init__(self, maxInitTime=10):                                          # StartServer
        """Starts the server, verify that it starts okay and dumps meta information (including
function signatures) to ``cbs.l``

:param maxInitTime: time to wait in seconds until the server is online before declaring it's unsuccessful""" # StartServer
        super().__init__(); self.maxInitTime = maxInitTime                       # StartServer
    def startServer(self):                                                       # StartServer
        pythonFile = self.l["pythonFile"]; metaFile = self.l["metaFile"]; port = self.l["port"]; maxInitTime = self.maxInitTime # StartServer
        node = self.l["node"]; condaEnv = self.l["condaEnv"]; condaPath = self.l["condaPath"]; startTime = time.time() # StartServer
        # print(f"{pythonFile=} {metaFile=} {port=} {maxInitTime=} {node=} {condaEnv=} {condaPath=}") # StartServer
        if node is None:                                                         # StartServer
            None | cli.cmd(f"python {pythonFile} &"); count = 0                  # StartServer
            while not os.path.exists(metaFile):                                  # StartServer
                if time.time()-startTime > maxInitTime: raise Exception(f"Tried to start server up, but no responses yet. Port: {port}, pythonFile: {pythonFile}, metaFile: {metaFile}") # StartServer
                time.sleep(0.1)                                                  # StartServer
            self.l["meta"] = meta = metaFile | cli.cat(text=False) | cli.aS(dill.loads) # StartServer
        else:                                                                    # StartServer
            if condaEnv: None | cli.cmd(f"ssh {node} 'source {condaPath}/bin/activate {condaEnv} && nohup python {pythonFile}' &"); # StartServer
            else: None | cli.cmd(f"ssh {node} 'nohup python {pythonFile}' &");   # StartServer
            while not int(None | cli.cmd(f"ssh {node} 'if [ -e {metaFile} ]; then echo 1; else echo 0; fi'") | cli.item()): # StartServer
                if time.time()-startTime > maxInitTime: raise Exception(f"Tried to start server up, but no responses yet. Port: {port}, pythonFile: {pythonFile}, metaFile: {metaFile}") # StartServer
                time.sleep(0.5)                                                  # StartServer
            self.l["meta"] = meta = dill.loads(b"".join(None | cli.cmd(f"ssh {node} 'cat {metaFile}'", text=False))) # StartServer
class GenerateHtml(k1.Callback):                                                 # GenerateHtml
    def __init__(self, serverPrefix=None, htmlFile=None, title="Interactive demo"): # GenerateHtml
        """Generates a html file that communicates with the server.

:param serverPrefix: prefix of server for back and forth requests, like "https://example.com/proj1". If
    empty, tries to grab ``cbs.l["serverPrefix"]``, which you can deposit from your own callback. If
    that's not available then it will fallback to ``localhost:port``
:param htmlFile: path of the target html file. If not specified then a temporary file
    will be created and made available in ``cbs.l["htmlFile"]``
:param title: title of html page"""                                              # GenerateHtml
        super().__init__(); self.serverPrefix = serverPrefix; self.htmlFile = htmlFile; self.title = title # GenerateHtml
    def generateHtml(self):                                                      # GenerateHtml
        meta = dict(self.l["meta"])                                              # GenerateHtml
        meta["fetch"]["js"] = meta["fetch"]["js"].replace("PREFIX", self.l["prefix"] or "PREFIX_") # GenerateHtml
        meta["fetch"]["py"] = meta["fetch"]["py"].replace("PREFIX", self.l["prefix"] or "PREFIX_") # GenerateHtml
        meta["fetch"]["k1"] = meta["fetch"]["k1"].replace("PREFIX", self.l["prefix"] or "PREFIX_") # GenerateHtml
        replaceNewlineWithBr = op().split("<!-- k1lib.raw.start -->") | apply(op().split("<!-- k1lib.raw.end -->")) | head(1).split() | (op().replace("\n", "<br>").all(2)) + apply(op().replace("\n", "<br>"), 1) | joinSt(2) | join("") # GenerateHtml
        replaces = cli.op().replace("META_JSON", base64.b64encode(_json.dumps(meta).encode()).decode())\
            .replace("SERVER_PREFIX", self.serverPrefix or self.l["serverPrefix"] or f"http://localhost:{self.l['port']}")\
            .replace("TITLE", self.title).replace("INTRO", meta["mainDoc"] | replaceNewlineWithBr | op().replace("\ue157", "\n"))\
            .replace("SOURCE_CODE", pretty_py(meta["source"]))                   # GenerateHtml
        self.l["htmlFile"] = cli.cat(f"{basePath}main.html") | replaces.all() | cli.file(self.htmlFile) # GenerateHtml
def commonCbs():                                                                 # commonCbs
    """Grabs common callbacks, including :class:`BuildPythonFile` and :class:`StartServer`""" # commonCbs
    return k1.Callbacks().add(BuildPythonFile()).add(StartServer());             # commonCbs
def serve(cbs):                                                                  # serve
    """Runs the serving pipeline."""                                             # serve
    import flask, flask_cors                                                     # serve
    cbs.l = defaultdict(lambda: None)                                            # serve
    cbs("begin")                                                                 # serve
    cbs("fetchSource") # fetches cells                                           # serve
    cbs("beforeBuildPythonFile"); cbs("buildPythonFile") # builds python server file # serve
    cbs("beforeStartServer"); cbs("startServer") # starts serving the model on localhost and add more meta info # serve
    cbs("beforeGenerateHtml"); cbs("generateHtml") # produces a standalone html file that provides interactive functionalities # serve
    cbs("end")                                                                   # serve
    return cbs                                                                   # serve
def cutoffLongText(s):                                                           # cutoffLongText
    if len(s) > 50:                                                              # cutoffLongText
        if isinstance(s, str):   return s[:50]+ "..."+s[-10:]                    # cutoffLongText
        if isinstance(s, bytes): return s[:50]+b"..."+s[-10:]                    # cutoffLongText
    return s                                                                     # cutoffLongText
def fetch_js(meta):                                                              # fetch_js
    a = meta["annos"].items() | filt(op()!="return", 0) | ~apply(lambda x,y: [x,y,meta["defaults"][x]]) | deref() # fetch_js
    ans = []; ret = meta["annos"]["return"]                                      # fetch_js
    for arg, anno, default in a:                                                 # fetch_js
        if anno in ("int", "float", "slider", "date", "text"): ans.append(f"    '{arg}': {_json.dumps(default[0])}") # fetch_js
        elif anno in ("apiKey",):     ans.append(f"    '{arg}': '<api key here>'") # fetch_js
        elif anno in ("checkbox",):   ans.append(f"    '{arg}': {_json.dumps(default)}") # fetch_js
        elif anno in ("json",):       ans.append(f"    '{arg}': {_json.dumps(default[0])} /* raw json object, no need to do JSON.stringify() */") # fetch_js
        elif anno in ("bytes",):      ans.append(f"    '{arg}': {cutoffLongText(_json.dumps(default))} /* base64-encoded bytes */") # fetch_js
        elif anno in ("image",):      ans.append(f"    '{arg}': {cutoffLongText(_json.dumps(default))} /* base64-encoded bytes of image in popular formats like jpg, png */") # fetch_js
        elif anno in ("serialized",): ans.append(f"    '{arg}': {cutoffLongText(_json.dumps(default))} /* base64-encoded bytes, encoded by python module `dill` */") # fetch_js
        elif anno in ("dropdown",):   ans.append(f"    '{arg}': {_json.dumps(default[1][default[0]])}") # fetch_js
    b = ",\n".join(ans)                                                          # fetch_js
    s = f"""
let result = (await (await fetch("https://local.mlexps.com/routeServer/PREFIX", {{
  method: "POST",
  body: JSON.stringify({{
{b}
  }}),
  headers: {{
    "Content-Type": "application/json",
  }}
}})).json()).data;"""                                                            # fetch_js
    if ret in ("int", "float"): s = f"{s} // returns {ret}"                      # fetch_js
    elif ret in ("text", "html"): s = f"{s} // returns string"                   # fetch_js
    elif ret in ("json",): s = f"{s} // returns js object, no need to do JSON.parse()" # fetch_js
    elif ret in ("bytes", "serialized"): s = f"{s} // returns base64-encoded bytes" # fetch_js
    elif ret in ("image",): s = f"{s} // returns base64-encoded bytes of an image, typically in jpg" # fetch_js
    return s                                                                     # fetch_js
def fetch_py(meta):                                                              # fetch_py
    a = meta["annos"].items() | filt(op()!="return", 0) | ~apply(lambda x,y: [x,y,meta["defaults"][x]]) | deref() # fetch_py
    ans = []; ret = meta["annos"]["return"]                                      # fetch_py
    for arg, anno, default in a:                                                 # fetch_py
        if anno in ("int", "float", "slider", "date", "text"): ans.append(f"    '{arg}': {_json.dumps(default[0])},") # fetch_py
        elif anno in ("apiKey",):     ans.append(f"    '{arg}': '<api key here>',") # fetch_py
        elif anno in ("checkbox",):   ans.append(f"    '{arg}': {default},")     # fetch_py
        elif anno in ("json",):       ans.append(f"    '{arg}': {default[0]}, # raw python object, no need to do json.dumps()") # fetch_py
        elif anno in ("bytes",):      ans.append(f"    '{arg}': base64.b64encode({cutoffLongText(base64.b64decode(default))}).decode(), # base64-encoded bytes") # fetch_py
        elif anno in ("image",):      ans.append(f"    '{arg}': base64.b64encode({cutoffLongText(base64.b64decode(default))}).decode(), # base64-encoded bytes of image in popular formats like jpg, png") # fetch_py
        elif anno in ("serialized",): ans.append(f"    '{arg}': base64.b64encode(dill.dumps(<raw_python_object_here>)).decode(), # base64-encoded bytes, encoded by python module `dill`") # fetch_py
        elif anno in ("dropdown",):   ans.append(f"    '{arg}': {_json.dumps(default[1][default[0]])},") # fetch_py
    b = "\n".join(ans)                                                           # fetch_py
    s = f"""
import requests, base64, dill, json
result = requests.post("https://local.mlexps.com/routeServer/PREFIX", json={{
{b}
}}).json()["data"]"""                                                            # fetch_py
    if ret in ("int", "float"): s = f"{s} # returns {ret}"                       # fetch_py
    elif ret in ("text", "html"): s = f"{s} # returns string"                    # fetch_py
    elif ret in ("json",): s = f"{s} # returns js object, no need to do json.loads()" # fetch_py
    elif ret in ("bytes",): s = f"{s} # returns base64-encoded bytes\nresult = base64.b64decode(result.encode())" # fetch_py
    elif ret in ("serialized",): s = f"{s} # returns base64-encoded bytes\nresult = dill.loads(base64.b64decode(result.encode()))" # fetch_py
    elif ret in ("image",): s = f"{s} # returns base64-encoded bytes of an image, typically in jpg" # fetch_py
    return s                                                                     # fetch_py
def fetch_k1(meta):                                                              # fetch_k1
    a = meta["annos"].items() | filt(op()!="return", 0) | ~apply(lambda x,y: [x,y,meta["defaults"][x]]) | deref() # fetch_k1
    ans = []; ret = meta["annos"]["return"]                                      # fetch_k1
    for arg, anno, default in a:                                                 # fetch_k1
        if anno in ("int", "float", "slider", "date", "text"): ans.append(f"    '{arg}': {_json.dumps(default[0])},") # fetch_k1
        elif anno in ("apiKey",):     ans.append(f"    '{arg}': '<api key here>',") # fetch_k1
        elif anno in ("checkbox",):   ans.append(f"    '{arg}': {default},")     # fetch_k1
        elif anno in ("json",):       ans.append(f"    '{arg}': {default[0]},")  # fetch_k1
        elif anno in ("bytes",):      ans.append(f"    '{arg}': {cutoffLongText(base64.b64decode(default))},") # fetch_k1
        elif anno in ("image",):      ans.append(f"    '{arg}': 'some/image/path.jpg' | toImg(), # raw PIL image") # fetch_k1
        elif anno in ("serialized",): ans.append(f"    '{arg}': <raw_python_object_here>,") # fetch_k1
        elif anno in ("dropdown",):   ans.append(f"    '{arg}': {_json.dumps(default[1][default[0]])},") # fetch_k1
    b = "\n".join(ans)                                                           # fetch_k1
    s = f"""{{\n{b}\n}} | kapi.demo("PREFIX")"""                                 # fetch_k1
    if ret in ("int", "float"): s = f"{s} # returns {ret}"                       # fetch_k1
    elif ret in ("text", "html"): s = f"{s} # returns string"                    # fetch_k1
    elif ret in ("json",): s = f"{s} # returns js object, no need to do json.loads()" # fetch_k1
    elif ret in ("bytes",): s = f"{s} # returns raw bytes, no need to base64-decode" # fetch_k1
    elif ret in ("serialized",): s = f"{s} # returns raw Python object, no need to base64-decode and dill-decode" # fetch_k1
    elif ret in ("image",): s = f"{s} # returns PIL image, no need to base64-decode" # fetch_k1
    return s                                                                     # fetch_k1
class baseType:                                                                  # baseType
    def __init__(self):                                                          # baseType
        """Base type for all widget types"""                                     # baseType
        pass                                                                     # baseType
    def getConfig(self): return NotImplemented                                   # baseType
class text(baseType):                                                            # text
    def __init__(self, multiline:bool=True, password:bool=False):                # text
        """Represents text, either on single or multiple lines.
If `password` is true, then will set multiline to false automatically,
and creates a text box that blurs out the contents. Example::

    def endpoint(s:serve.text()="abc") -> str: pass

For inputs only. Use ``str`` for outputs"""                                      # text
        super().__init__(); self.multiline = multiline if not password else False; self.password = password # text
    def __repr__(self): return f"<text multiline={self.multiline}>"              # text
class slider(baseType):                                                          # slider
    def __init__(self, start:float, stop:float, intervals:int=100):              # slider
        """Represents a slider from `start` to `stop` with a bunch of
intervals in the middle. If `defValue` is not specified, uses the
middle point between start and stop. Example::

    def endpoint(a:serve.slider(2, 3.2)=2.3) -> str: pass

For inputs only"""                                                               # slider
        super().__init__(); self.start = start; self.stop = stop; self.intervals = intervals; self.dt = (stop-start)/intervals; self.step = self.dt # slider
    def __repr__(self): return f"<slider {self.start}---{self.intervals}-->{self.stop}>" # slider
class html(baseType):                                                            # html
    def __init__(self):                                                          # html
        """Raw html.
Example::

    def endpoint() -> serve.html(): pass

For outputs only"""                                                              # html
        super().__init__()                                                       # html
    def __repr__(self): return f"<html>"                                         # html
class json(baseType):                                                            # json
    def __init__(self):                                                          # json
        """Raw json.
Example::

    def endpoint(a:serve.json()={"a": 3}) -> serve.json(): pass

For inputs and outputs"""                                                        # json
        super().__init__()                                                       # json
    def __repr__(self): return f"<json>"                                         # json
class date(baseType):                                                            # date
    def __init__(self, min=None, max=None):                                      # date
        """Local date time (no timezone information).
Example::

    def endpoint(d:serve.date()="2023-12-07T00:00:00") -> str: pass

:param min: min date, also in format '2023-12-07T00:00:00'"""                    # date
        super().__init__(); self.minDate = min; self.maxDate = max               # date
    def __repr__(self): return f"<date>"                                         # date
class serialized(baseType):                                                      # serialized
    def __init__(self):                                                          # serialized
        """For serialized objects using :mod:`dill`.
Example::

    def endpoint(a:serve.serialized()) -> serve.serialized():
        return {"any": "data structure", "you": "want", "even": np.random.randn(100)}
"""                                                                              # serialized
        super().__init__()                                                       # serialized
    def __repr__(self): return f"<serialized>"                                   # serialized
class apiKey(baseType):                                                          # apiKey
    def __init__(self, apiKey=str):                                              # apiKey
        """Protects your endpoint with an api key.
Example::

    def endpoint(apiKey:serve.apiKey("your api key here")="") -> str: pass

When compiled, your api key won't appear anywhere, not in the html, not in the meta
files, and someone calling the endpoint must specify it, else it will just errors out""" # apiKey
        super().__init__(); self.apiKey = apiKey                                 # apiKey
    def __repr__(self): return f"<apiKey>"                                       # apiKey
def refine(param:str, anno:baseType, default): # anno is not necessarily baseType, can be other types like "int" # refine
    if anno == int: return [param, "int", [default, False], None]                # refine
    if anno == float: return [param, "float", [default, False], None]            # refine
    multiline = lambda s: len(s.split("\n")) > 1 or len(s) > 100                 # refine
    if anno == bool: return [param, "checkbox", default, None]                   # refine
    if anno == str: return [param, "text", [default, multiline(default or "")], None] # refine
    if isinstance(anno, text): return [param, "text", [default, anno.multiline, anno.password], None] # refine
    if isinstance(anno, slider): return [param, "slider", [default, anno.start, anno.stop, anno.dt], None] # refine
    if isinstance(anno, range): return [param, "slider", [default, anno.start, anno.stop, anno.step], None] # refine
    byte2Str = aS(base64.b64encode) | op().decode("ascii")                       # refine
    if hasPIL and anno == PIL.Image.Image: return [param, "image", (default | toBytes() | byte2Str) if default is not None else None, None] # refine
    if anno == bytes: return [param, "bytes", (default | byte2Str) if default is not None else None, None] # refine
    if isinstance(anno, serialized): return [param, "serialized", (default | aS(dill.dumps) | byte2Str) if default is not None else None, None] # refine
    if isinstance(anno, list): anno | apply(str) | deref(); return [param, "dropdown", [anno.index(default), anno], None] # refine
    if isinstance(anno, html): return [param, "html", [default], None]           # refine
    if isinstance(anno, json): return [param, "json", [default, True], None]     # refine
    if isinstance(anno, date): return [param, "date", [default, anno.minDate, anno.maxDate], None] # refine
    if isinstance(anno, apiKey): return [param, "apiKey", [default], anno.apiKey] # refine
    raise Exception(f"Unknown type {anno}")                                      # refine
def analyze(f):                                                                  # analyze
    spec = getattr(f, "fullargspec", inspect.getfullargspec(f)); args = spec.args; n = len(args) # analyze
    annos = spec.annotations; defaults = spec.defaults or ()                     # analyze
    docs = (f.__doc__ or "").split("\n") | grep(":param", sep=True).till() | filt(op().ab_len() > 0) | op().strip().all(2) | (join(" ") | op().split(":") | ~aS(lambda x, y, *z: [y.split(" ")[1], ":".join(z).strip()])).all() | toDict() # analyze
    mainDoc = (f.__doc__ or " ").split("\n") | grep(".", sep=True).till(":param") | breakIf(op()[0].startswith(":param")) | join("\n").all() | join("\n") # analyze
                                                                                 # analyze
    if len(annos) != n + 1: raise Exception(f"Please annotate all of your arguments ({n} args + 1 return != {len(annos)} annos). Args: {args}, annos: {annos}") # analyze
    if len(defaults) != n: raise Exception(f"Please specify default values for all of your arguments ({n} args != {len(defaults)} default values)") # analyze
    a = [args, args | lookup(annos), defaults] | transpose() | ~apply(refine) | deref() # analyze
    ret = refine("return", annos["return"], None)[1]; defaults = a | cut(0, 2) | toDict() # analyze
    annos = a | cut(0, 1) | toDict(); annos["return"] = ret; privates = a | cut(0, 3) | toDict() # analyze
    if ret == "slider": raise Exception(f"Return value is a slider, which doesn't really make sense") # analyze
    if ret == "date": raise Exception(f"Return value is a date, which doesn't really make sense. Use `str` instead") # analyze
    if ret == "apiKey": raise Exception(f"Return value is an api key, which doesn't really make sense") # analyze
    if ret == "checkbox": raise Exception(f"Return value is a checkbox, which doesn't really make sense. Use `serve.json()` instead") # analyze
    for anno in [v for k,v in annos.items() if k != "return"]:                   # analyze
        if anno == "html": raise Exception("One of the input arguments is serve.html(), which doesn't really make sense. Use serve.text() instead") # analyze
                                                                                 # analyze
    # args:list, annos:dict, defaults:list, docs:dict                            # analyze
                                                                                 # analyze
    meta = {"args": args, "annos": annos, "defaults": defaults, "privates": privates, "docs": docs, # analyze
     "mainDoc": mainDoc, "source": inspect.getsource(f), "pid": os.getpid()}     # analyze
    f_js = pretty_js(fetch_js(meta)); f_py = pretty_py(fetch_py(meta)); f_k1 = pretty_py(fetch_k1(meta)) # analyze
    return {**meta, "fetch": {"js": f_js, "py": f_py, "k1": f_k1}}               # analyze
class Html(str):                                                                 # Html
    def _repr_html_(self): return self                                           # Html
def webToPy(o, klass:baseType):                                                  # webToPy
    if klass == "int": return int(o)                                             # webToPy
    if klass in ("float", "text", "checkbox", "dropdown", "slider", "apiKey", "date", "json"): return o # webToPy
    o = str(o)                                                                   # webToPy
    if klass == "bytes": return base64.b64decode(o)                              # webToPy
    if klass == "serialized": return dill.loads(base64.b64decode(o))             # webToPy
    if klass == "image": return o | aS(base64.b64decode) | toImg()               # webToPy
    if klass == "html": return Html(base64.b64decode(o).decode())                # webToPy
    return NotImplemented                                                        # webToPy
def pyToWeb(o, klass:baseType):                                                  # pyToWeb
    if klass in ("int", "float", "text", "checkbox", "slider", "apiKey", "date", "json"): return o # pyToWeb
    if klass == "bytes": return base64.b64encode(o).decode()                     # pyToWeb
    if klass == "serialized": return base64.b64encode(dill.dumps(o)).decode()    # pyToWeb
    if klass == "image": return base64.b64encode(o | toBytes()).decode()         # pyToWeb
    if klass == "dropdown": return o;                                            # pyToWeb
    if klass == "html": return o.encode() | aS(base64.b64encode) | op().decode() # pyToWeb
    return NotImplemented                                                        # pyToWeb



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .callbacks import Callback, Callbacks, Cbs
import k1lib, os, warnings
from typing import Callable
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["Autosave", "DontTrainValid", "InspectLoss", "ModifyLoss", "Cpu", "Cuda",
           "DType", "InspectBatch", "ModifyBatch", "InspectOutput", "ModifyOutput",
           "Beep", "OnProgress"]
@k1lib.patch(Cbs)
class Autosave(Callback):                                                        # Autosave
    """Autosaves 3 versions of the network to disk"""                            # Autosave
    def __init__(self): super().__init__(); self.order = 23                      # Autosave
    def endRun(self):                                                            # Autosave
        os.system("mv autosave-1.pth autosave-0.pth")                            # Autosave
        os.system("mv autosave-2.pth autosave-1.pth")                            # Autosave
        self.l.save("autosave-2.pth")                                            # Autosave
@k1lib.patch(Cbs)                                                                # Autosave
class DontTrainValid(Callback):                                                  # DontTrainValid
    """If is not training, then don't run m.backward() and opt.step().
The core training loop in k1lib.Learner don't specifically do this,
cause there may be some weird cases where you want to also train valid."""       # DontTrainValid
    def _common(self):                                                           # DontTrainValid
        if not self.l.model.training: return True                                # DontTrainValid
    def startBackward(self): return self._common()                               # DontTrainValid
    def startStep(self): return self._common()                                   # DontTrainValid
@k1lib.patch(Cbs)                                                                # DontTrainValid
class InspectLoss(Callback):                                                     # InspectLoss
    """Expected `f` to take in 1 float."""                                       # InspectLoss
    def __init__(self, f): super().__init__(); self.f = f; self.order = 15       # InspectLoss
    def endLoss(self): self.f(self.loss.detach())                                # InspectLoss
@k1lib.patch(Cbs)                                                                # InspectLoss
class ModifyLoss(Callback):                                                      # ModifyLoss
    """Expected `f` to take in 1 float and return 1 float."""                    # ModifyLoss
    def __init__(self, f): super().__init__(); self.f = f; self.order = 13       # ModifyLoss
    def endLoss(self): self.l.loss = self.f(self.loss)                           # ModifyLoss
@k1lib.patch(Cbs)                                                                # ModifyLoss
class Cuda(Callback):                                                            # Cuda
    """Moves batch and model to the default GPU"""                               # Cuda
    def startRun(self): self.l.model.cuda()                                      # Cuda
    def startBatch(self):                                                        # Cuda
        try: self.l.xb = self.l.xb.cuda()                                        # Cuda
        except Exception as e: warnings.warn(f"xb can't be moved to the GPU: {e}") # Cuda
        try: self.l.yb = self.l.yb.cuda()                                        # Cuda
        except Exception as e: warnings.warn(f"yb can't be moved to the GPU: {e}") # Cuda
@k1lib.patch(Cbs)                                                                # Cuda
class Cpu(Callback):                                                             # Cpu
    """Moves batch and model to CPU"""                                           # Cpu
    def startRun(self): self.l.model.cpu()                                       # Cpu
    def startBatch(self):                                                        # Cpu
        self.l.xb = self.l.xb.cpu()                                              # Cpu
        self.l.yb = self.l.yb.cpu()                                              # Cpu
@k1lib.patch(Cbs)                                                                # Cpu
class DType(Callback):                                                           # DType
    """Moves batch and model to a specified data type"""                         # DType
    def __init__(self, dtype): super().__init__(); self.dtype = dtype            # DType
    def startRun(self): self.l.model = self.l.model.to(self.dtype)               # DType
    def startBatch(self):                                                        # DType
        self.l.xb = self.l.xb.to(self.dtype)                                     # DType
        self.l.yb = self.l.yb.to(self.dtype)                                     # DType
@k1lib.patch(Cbs)                                                                # DType
class InspectBatch(Callback):                                                    # InspectBatch
    """Expected `f` to take in 2 tensors."""                                     # InspectBatch
    def __init__(self, f:callable): super().__init__(); self.f = f; self.order = 15 # InspectBatch
    def startBatch(self): self.f(self.l.xb, self.l.yb)                           # InspectBatch
@k1lib.patch(Cbs)                                                                # InspectBatch
class ModifyBatch(Callback):                                                     # ModifyBatch
    """Modifies xb and yb on the fly. Expected `f`
    to take in 2 tensors and return 2 tensors."""                                # ModifyBatch
    def __init__(self, f): super().__init__(); self.f = f; self.order = 13       # ModifyBatch
    def startBatch(self): self.l.xb, self.l.yb = self.f(self.l.xb, self.l.yb)    # ModifyBatch
@k1lib.patch(Cbs)                                                                # ModifyBatch
class InspectOutput(Callback):                                                   # InspectOutput
    """Expected `f` to take in 1 tensor."""                                      # InspectOutput
    def __init__(self, f): super().__init__(); self.f = f; self.order = 15       # InspectOutput
    def endPass(self): self.f(self.y)                                            # InspectOutput
@k1lib.patch(Cbs)                                                                # InspectOutput
class ModifyOutput(Callback):                                                    # ModifyOutput
    """Modifies output on the fly. Expected `f` to take
in 1 tensor and return 1 tensor"""                                               # ModifyOutput
    def __init__(self, f): super().__init__(); self.f = f; self.order = 13       # ModifyOutput
    def endPass(self): self.l.y = self.f(self.y)                                 # ModifyOutput
@k1lib.patch(Cbs)                                                                # ModifyOutput
class Beep(Callback):                                                            # Beep
    """Plays a beep sound when the run is over"""                                # Beep
    def endRun(self): k1lib.beep()                                               # Beep
@k1lib.patch(Cbs)                                                                # Beep
class OnProgress(Callback):                                                      # OnProgress
    """Triggers a specific function once when reached a certain
progress.

:param f: function that takes in a :class:`~k1lib.Learner` object
:param progress: progress to trigger the function, from 0 to 1"""                # OnProgress
    def __init__(self, f:Callable[["k1lib.Learner"], None], progress:float):     # OnProgress
        super().__init__(); self.f = f; self.progress = progress; self.ran = False # OnProgress
    def startRun(self): self.ran = False                                         # OnProgress
    def startBatch(self):                                                        # OnProgress
        if (not self.ran) and self.l.progress > self.progress:                   # OnProgress
            self.ran = True; self.f(self.l)                                      # OnProgress



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This module is for selecting a subnetwork using CSS so that you can do special
things to them. Checkout the tutorial section for a walkthrough. This is exposed
automatically with::

   from k1lib.imports import *
   selector.select # exposed
"""
import k1lib, re
from k1lib import cli
from typing import List, Tuple, Dict, Union, Any, Iterator, Callable
from contextlib import contextmanager; from functools import partial
try: import torch; from torch import nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["ModuleSelector", "preprocess", "select"]
def preprocess(selectors:str, defaultProp="*") -> List[str]:                     # preprocess
    r"""Removes all quirkly features allowed by the css
language, and outputs nice lines. Example::

    # returns ["a:f", "a:g,h", "b:g,h", "t:*"]
    selector.preprocess("a:f; a, b: g,h; t")

:param selectors: single css selector string. Statements separated
    by "\\n" or ";"
:param defaultProp: default property, if statement doesn't have one"""           # preprocess
    # filtering unwanted characters and quirky spaces                            # preprocess
    lines = [e for l in selectors.split("\n") for e in l.split(";")]             # preprocess
    selectors = [re.sub("(^\s+)|(\s+$)", "", re.sub("\s\s+", " ", line)).replace(" >", ">").replace("> ", ">").replace(" :", ":").replace(": ", ":").replace(" ,", ",").replace(", ", ",").replace(";", "\n").replace(" \n", "\n").replace("\n ", "\n") for line in lines if line != ""] # preprocess
    # adding "*" to all selectors with no props specified                        # preprocess
    selectors = [selector if ":" in selector else f"{selector}:{defaultProp}" for selector in selectors] # preprocess
    # expanding comma-delimited selectors                                        # preprocess
    return [f"{segment}:{selector.split(':')[1]}" for selector in selectors for segment in selector.split(":")[0].split(",")] # preprocess
def _getParts(s:str): return [a for elem in s.split(":")[0].split(">") if elem for a in elem.split(" ") if a] # _getParts
def _getProps(s:str): return [elem for elem in s.split(":")[1].split(",") if elem] # _getProps
_idxAuto = k1lib.AutoIncrement()                                                 # _getProps
class ModuleSelector: # empty methods so that Sphinx generates the docs in order # ModuleSelector
    props:List[str]                                                              # ModuleSelector
    """Properties of this :class:`ModuleSelector`"""                             # ModuleSelector
    idx:int                                                                      # ModuleSelector
    """Unique id of this :class:`ModuleSelector` in the entire script. May be useful
for module recognition"""                                                        # ModuleSelector
    nn:"torch.nn.Module"                                                         # ModuleSelector
    """The associated :class:`torch.nn.Module` of this :class:`ModuleSelector`""" # ModuleSelector
    def __init__(self, parent:"ModuleSelector", name:str, nn:"torch.nn.Module"): # ModuleSelector
        self.parent = parent; self.name = name; self.nn = nn                     # ModuleSelector
        self._children:Dict["ModuleSelector"] = {}                               # ModuleSelector
        self.props:List[str] = []; self.depth:int = 0                            # ModuleSelector
        self.directSelectors:List[str] = []                                      # ModuleSelector
        self.indirectSelectors:List[str] = []                                    # ModuleSelector
        self.displayF:Callable[["ModuleSelector"], str] = lambda mS: ', '.join(mS.props) # ModuleSelector
        self.idx = _idxAuto()                                                    # ModuleSelector
    def deepestDepth(self): pass                                                 # ModuleSelector
    def highlight(self, prop:str):                                               # ModuleSelector
        """Highlights the specified prop when displaying the object."""          # ModuleSelector
        self.displayF = lambda self: (k1lib.fmt.txt.red if prop in self else k1lib.fmt.txt.identity)(', '.join(self.props)) # ModuleSelector
        return self                                                              # ModuleSelector
    def __call__(self, *args, **kwargs):                                         # ModuleSelector
        """Calls the internal :class:`torch.nn.Module`"""                        # ModuleSelector
        return self.nn(*args, **kwargs)                                          # ModuleSelector
    def __contains__(self): pass                                                 # ModuleSelector
    def named_children(self): pass                                               # ModuleSelector
    def children(self): pass                                                     # ModuleSelector
    def named_modules(self): pass                                                # ModuleSelector
    def modules(self): pass                                                      # ModuleSelector
    def directParams(self): pass                                                 # ModuleSelector
    def parse(self): pass                                                        # ModuleSelector
    def apply(self): pass                                                        # ModuleSelector
    def clearProps(self): pass                                                   # ModuleSelector
    @property                                                                    # ModuleSelector
    def displayF(self):                                                          # ModuleSelector
        """Function to display each ModuleSelector's lines.
Default is just::

    lambda mS: ", ".join(mS.props) """                                           # ModuleSelector
        return self._displayF                                                    # ModuleSelector
    @displayF.setter                                                             # ModuleSelector
    def displayF(self, f):                                                       # ModuleSelector
        def applyF(self): self._displayF = f                                     # ModuleSelector
        self.apply(applyF)                                                       # ModuleSelector
    def __getattr__(self, attr):                                                 # ModuleSelector
        if attr.startswith("_"): raise AttributeError(attr)                      # ModuleSelector
        if attr in self._children: return self._children[attr]                   # ModuleSelector
        return self.directParams[attr]                                           # ModuleSelector
    def __getitem__(self, idx): return getattr(self, str(idx))                   # ModuleSelector
    @staticmethod                                                                # ModuleSelector
    def sample() -> "ModuleSelector":                                            # ModuleSelector
        """Create a new example :class:`ModuleSelector` that has a bit of
hierarchy to them, with no css."""                                               # ModuleSelector
        return nn.Sequential(nn.Linear(3, 4), nn.Sequential(nn.Conv2d(3, 8, 3, 2), nn.ReLU(), nn.Linear(5, 6)), nn.Linear(7, 8)).select("") # ModuleSelector
    def hookF(self): pass                                                        # ModuleSelector
    def hookFp(self): pass                                                       # ModuleSelector
    def hookB(self): pass                                                        # ModuleSelector
    def freeze(self): pass                                                       # ModuleSelector
    def unfreeze(self): pass                                                     # ModuleSelector
@k1lib.patch(nn.Module)                                                          # ModuleSelector
def select(model:"torch.nn.Module", css:str="*") -> "k1lib.selector.ModuleSelector": # select
    """Creates a new ModuleSelector, in sync with a model.
Example::

    mS = selector.select(nn.Linear(3, 4), "#root:propA")

Or, you can do it the more direct way::

    mS = nn.Linear(3, 4).select("#root:propA")

:param model: the :class:`torch.nn.Module` object to select from
:param css: the css selectors"""                                                 # select
    root = ModuleSelector(None, "root", model)                                   # select
    root.parse(preprocess(css)); return root                                     # select
@k1lib.patch(ModuleSelector, name="apply")                                       # select
def _apply(self, f:Callable[[ModuleSelector], None]):                            # _apply
    """Applies a function to self and all child :class:`ModuleSelector`"""       # _apply
    f(self)                                                                      # _apply
    for child in self._children.values(): child.apply(f)                         # _apply
@k1lib.patch(ModuleSelector, name="parse")                                       # _apply
def _parse(self, selectors:Union[List[str], str]) -> ModuleSelector:             # _parse
    """Parses extra selectors. Clears all old selectors, but retain
the props. Returns self. Example::

    mS = selector.ModuleSelector.sample().parse("Conv2d:propA")
    # returns True
    "propA" in mS[1][0]

:param selectors: can be the preprocessed list, or the unprocessed css string""" # _parse
    if isinstance(selectors, str): selectors = preprocess(selectors)             # _parse
    self.directSelectors = []; self.indirectSelectors = []                       # _parse
    ogSelectors = selectors                                                      # _parse
    if self.parent != None:                                                      # _parse
        selectors = [] + selectors + self.parent.indirectSelectors + self.parent.directSelectors # _parse
        self.indirectSelectors += self.parent.indirectSelectors                  # _parse
        self.depth = self.parent.depth + 1                                       # _parse
    for selector in selectors:                                                   # _parse
        parts = _getParts(selector)                                              # _parse
        matches = parts[0] == self.nn.__class__.__name__ or parts[0] == "#" + self.name or parts[0] == "*" # _parse
        if len(parts) == 1:                                                      # _parse
            if matches: self.props += _getProps(selector)                        # _parse
        else:                                                                    # _parse
            a = selector.find(">"); a = a if a > 0 else float("inf")             # _parse
            b = selector.find(" "); b = b if b > 0 else float("inf")             # _parse
            direct = a < b                                                       # _parse
            if matches:                                                          # _parse
                if direct: self.directSelectors.append(selector[a+1:])           # _parse
                else: self.indirectSelectors.append(selector[b+1:])              # _parse
    for name, mod in self.nn.named_children():                                   # _parse
        if name not in self._children:                                           # _parse
            self._children[name] = ModuleSelector(self, name, mod)               # _parse
        self._children[name].parse(ogSelectors)                                  # _parse
    self.props = list(set(self.props)); return self                              # _parse
@k1lib.patch(ModuleSelector)                                                     # _parse
def __contains__(self, prop:str=None) -> bool:                                   # __contains__
    """Whether this :class:`ModuleSelector` has a specific prop.
Example::

    # returns True
    "b" in nn.Linear(3, 4).select("*:b")
    # returns False
    "h" in nn.Linear(3, 4).select("*:b")
    # returns True, "*" here means the ModuleSelector has any properties at all
    "*" in nn.Linear(3, 4).select("*:b")"""                                      # __contains__
    if "*" in self.props: return True                                            # __contains__
    if prop in self.props: return True                                           # __contains__
    if prop == "*" and len(self.props) > 0: return True                          # __contains__
    return False                                                                 # __contains__
@k1lib.patch(ModuleSelector)                                                     # __contains__
def named_children(self, prop:str=None) -> Iterator[Tuple[str, ModuleSelector]]: # named_children
    """Get all named direct childs.

:param prop: Filter property. See also: :meth:`__contains__`"""                  # named_children
    if prop is None: return self._children.items()                               # named_children
    return ((k, v) for k, v in self._children.items() if prop in v)              # named_children
@k1lib.patch(ModuleSelector)                                                     # named_children
def children(self, prop:str=None) -> Iterator[ModuleSelector]:                   # children
    """Get all direct childs.

:param prop: Filter property. See also: :meth:`__contains__`"""                  # children
    return (x for _, x in self.named_children(prop))                             # children
@k1lib.patch(ModuleSelector, "directParams")                                     # children
@property                                                                        # children
def directParams(self) -> Dict[str, nn.Parameter]:                               # directParams
    """Dict params directly under this module"""                                 # directParams
    return {name: param for name, param in self.nn.named_parameters() if "." not in name} # directParams
@k1lib.patch(ModuleSelector)                                                     # directParams
def named_modules(self, prop:str=None) -> Iterator[Tuple[str, ModuleSelector]]:  # named_modules
    """Get all named child recursively.
Example::

    modules = list(nn.Sequential(nn.Linear(3, 4), nn.ReLU()).select().named_modules())
    # return 3
    len(modules)
    # return tuple ('0', <ModuleSelector of Linear>)
    modules[1]

:param prop: Filter property. See also: :meth:`__contains__`"""                  # named_modules
    if prop != None:                                                             # named_modules
        yield from ((name, m) for name, m in self.named_modules() if prop in m)  # named_modules
        return                                                                   # named_modules
    yield self.name, self                                                        # named_modules
    for child in self._children.values(): yield from child.named_modules()       # named_modules
@k1lib.patch(ModuleSelector)                                                     # named_modules
def modules(self, prop:str=None) -> Iterator[ModuleSelector]:                    # modules
    """Get all child recursively.

:param prop: Filter property. See also: :meth:`__contains__`"""                  # modules
    for name, x in self.named_modules(prop): yield x                             # modules
@k1lib.patch(ModuleSelector)                                                     # modules
def clearProps(self) -> "ModuleSelector":                                        # clearProps
    """Clears all existing props of this and all descendants
:class:`ModuleSelector`. Example::

    # returns False
    "b" in nn.Linear(3, 4).select("*:b").clearProps()"""                         # clearProps
    def applyF(self): self.props = []                                            # clearProps
    self.apply(applyF); return self                                              # clearProps
@k1lib.patch(ModuleSelector, name="deepestDepth")                                # clearProps
@property                                                                        # clearProps
def deepestDepth(self):                                                          # deepestDepth
    """Deepest depth of the tree. If self doesn't
have any child, then depth is 0"""                                               # deepestDepth
    if len(self._children) == 0: return 0                                        # deepestDepth
    return 1 + max([child.deepestDepth for child in self._children.values()])    # deepestDepth
@k1lib.patch(ModuleSelector)                                                     # deepestDepth
def __repr__(self, intro:bool=True, header:Union[str, Tuple[str]]="", footer="", tabs:int=None): # __repr__
    """
    :param intro: whether to include a nice header and footer info
    :param header:
        str: include a header that starts where `displayF` will start
        Tuple[str, str]: first one in tree, second one in displayF section
    :param footer: same thing with header, but at the end
    :param header: include a header that starts where `displayF` will start
    :param tabs: number of tabs at the beginning. Best to leave this empty
    """                                                                          # __repr__
    if tabs == None: tabs = 5 + self.deepestDepth                                # __repr__
    answer = "ModuleSelector:\n" if intro else ""                                # __repr__
    if header:                                                                   # __repr__
        h1, h2 = ("", header) if isinstance(header, str) else header             # __repr__
        answer += h1.ljust(tabs*4, " ") + h2 + "\n"                              # __repr__
    answer += f"{self.name}: {self.nn.__class__.__name__}".ljust(tabs*4, " ")    # __repr__
    answer += self.displayF(self) + ("\n" if len(self._children) > 0 else "")    # __repr__
    answer += self._children.values() | cli.apply(lambda child: child.__repr__(tabs=tabs-1, intro=False).split("\n")) | cli.joinStreams() | cli.tab() | cli.join("\n") # __repr__
    if footer:                                                                   # __repr__
        f1, f2 = ("", footer) if isinstance(footer, str) else footer             # __repr__
        answer += "\n" + f1.ljust(tabs*4, " ") + f2                              # __repr__
    if intro: answer += f"""\n\nCan...
- mS.deepestDepth: get deepest depth possible
- mS.nn: get the underlying nn.Module object
- mS.apply(f): apply to self and all descendants
- "HookModule" in mS: whether this module has a specified prop
- mS.highlight(prop): highlights all modules with specified prop
- mS.parse([..., ...]): parses extra css
- mS.directParams: get Dict[str, nn.Parameter] that are directly under this module""" # __repr__
    return answer                                                                # __repr__
def _strTensor(t): return "None" if t is None else f"{t.shape}"                  # _strTensor
def strTensorTuple(ts):                                                          # strTensorTuple
    if len(ts) > 1:                                                              # strTensorTuple
        shapes = "\n".join(f"- {_strTensor(t)}" for t in ts)                     # strTensorTuple
        return f"tensors ({len(ts)} total) shapes:\n{shapes}"                    # strTensorTuple
    else:                                                                        # strTensorTuple
        return f"tensor shape: {_strTensor(ts[0])}"                              # strTensorTuple
@k1lib.patch(ModuleSelector)                                                     # strTensorTuple
@contextmanager                                                                  # strTensorTuple
def hookF(self, f:Callable[[ModuleSelector, "torch.nn.Module", Tuple[torch.Tensor], torch.Tensor], None]=None, prop:str="*"): # hookF
    """Context manager for applying forward hooks.
Example::

    def f(mS, i, o):
        print(i, o)

    m = nn.Linear(3, 4)
    with m.select().hookF(f):
        m(torch.randn(2, 3))

:param f: hook callback, should accept :class:`ModuleSelector`, inputs and output
:param prop: filter property of module to hook onto. If not specified, then it will print out input and output tensor shapes.""" # hookF
    if f is None: f = lambda mS, i, o: print(f"Forward hook {m}:\n" + ([f"Input  {strTensorTuple(i)}", f"Output tensor shape: {o.shape}"] | cli.tab() | cli.join("\n"))) # hookF
    g = lambda m, i, o: f(self, i, o)                                            # hookF
    handles = [m.nn.register_forward_hook(g) for m in self.modules(prop)]        # hookF
    try: yield                                                                   # hookF
    finally:                                                                     # hookF
        for h in handles: h.remove()                                             # hookF
@k1lib.patch(ModuleSelector)                                                     # hookF
@contextmanager                                                                  # hookF
def hookFp(self, f=None, prop:str="*"):                                          # hookFp
    """Context manager for applying forward pre hooks.
Example::

    def f(mS, i):
        print(i)

    m = nn.Linear(3, 4)
    with m.select().hookFp(f):
        m(torch.randn(2, 3))

:param f: hook callback, should accept :class:`ModuleSelector` and inputs
:param prop: filter property of module to hook onto. If not specified, then it will print out input tensor shapes.""" # hookFp
    if f is None: f = lambda mS, i: print(f"Forward pre hook {m}:\n" + ([f"Input {strTensorTuple(i)}"] | cli.tab() | cli.join("\n"))) # hookFp
    g = lambda m, i: f(self, i)                                                  # hookFp
    handles = [m.nn.register_forward_pre_hook(g) for m in self.modules(prop)]    # hookFp
    try: yield                                                                   # hookFp
    finally:                                                                     # hookFp
        for h in handles: h.remove()                                             # hookFp
@k1lib.patch(ModuleSelector)                                                     # hookFp
@contextmanager                                                                  # hookFp
def hookB(self, f=None, prop:str="*"):                                           # hookB
    """Context manager for applying backward hooks.
Example::

    def f(mS, i, o):
        print(i, o)

    m = nn.Linear(3, 4)
    with m.select().hookB(f):
        m(torch.randn(2, 3)).sum().backward()

:param f: hook callback, should accept :class:`ModuleSelector`, grad inputs and outputs
:param prop: filter property of module to hook onto. If not specified, then it will print out input tensor shapes.""" # hookB
    if f is None: f = lambda mS, i, o: print(f"Backward hook {m}:\n" + ([f"Input  {strTensorTuple(i)}", f"Output {strTensorTuple(o)}"] | cli.tab() | cli.join("\n"))) # hookB
    g = lambda m, i, o: f(self, i, o)                                            # hookB
    handles = [m.nn.register_full_backward_hook(g) for m in self.modules(prop)]  # hookB
    try: yield                                                                   # hookB
    finally:                                                                     # hookB
        for h in handles: h.remove()                                             # hookB
from contextlib import ExitStack                                                 # hookB
@contextmanager                                                                  # hookB
def _intercept(self, value:bool):                                                # _intercept
    handles = []                                                                 # _intercept
    try:                                                                         # _intercept
        data = []                                                                # _intercept
        f = lambda x: x.detach() if x is not None else None                      # _intercept
        for m in self.modules("*"):                                              # _intercept
            subData1 = []; subData2 = []; data.append([subData1, subData2])      # _intercept
            handles.append(m.nn.register_forward_hook(lambda _m, i, o: subData1.append([[f(e) for e in i], f(o)]))) # _intercept
            handles.append(m.nn.register_full_backward_hook(lambda _m, i, o: subData2.append([[f(e) for e in i], [f(e) for e in o]]))) # _intercept
        yield data                                                               # _intercept
    finally:                                                                     # _intercept
        for h in handles: h.remove()                                             # _intercept
@k1lib.patch(ModuleSelector)                                                     # _intercept
def intercept(self):                                                             # intercept
    """Returns a context manager that intercept forward and backward signals
to parts of the network. Example::

    l = k1lib.Learner.sample()
    with l.model.select("#lin1").intercept() as d:
        l.run(2)
    # returns (1, 2, 600, 2, 1, 32, 1), or (#selected modules, [forward, backward], #steps, [input, output], actual data)
    d | shape()

See also: :meth:`hookF`, :meth:`hookFp`, :meth:`hookB`"""                        # intercept
    return _intercept(self, False)                                               # intercept
from contextlib import ExitStack                                                 # intercept
@contextmanager                                                                  # intercept
def _freeze(self, value:bool, prop:str):                                         # _freeze
    with ExitStack() as stack:                                                   # _freeze
        for m in self.modules(prop):                                             # _freeze
            stack.enter_context(m.nn.gradContext())                              # _freeze
            m.nn.requires_grad_(value)                                           # _freeze
        try: yield                                                               # _freeze
        finally: pass                                                            # _freeze
@k1lib.patch(ModuleSelector)                                                     # _freeze
def freeze(self, prop:str="*"):                                                  # freeze
    """Returns a context manager that freezes (set requires_grad to False) parts of
the network. Example::

    l = k1lib.Learner.sample()
    w = l.model.lin1.lin.weight.clone() # weights before
    with l.model.select("#lin1").freeze():
        l.run(1)
    # returns True
    (l.model.lin1.lin.weight ==  w).all()"""                                     # freeze
    return _freeze(self, False, prop)                                            # freeze
@k1lib.patch(ModuleSelector)                                                     # freeze
def unfreeze(self, prop:str="*"):                                                # unfreeze
    """Returns a context manager that unfreezes (set requires_grad to True) parts of
the network. Example::

    l = k1lib.Learner.sample()
    w = l.model.lin1.lin.weight.clone() # weights before
    with l.model.select("#lin1").freeze():
        with l.model.select("#lin1 > #lin").unfreeze():
            l.run(1)
    # returns False
    (l.model.lin1.lin.weight ==  w).all()"""                                     # unfreeze
    return _freeze(self, True, prop)                                             # unfreeze
class CutOff(nn.Module):                                                         # CutOff
    def __init__(self, net, m):                                                  # CutOff
        super().__init__()                                                       # CutOff
        self.net = net; self.m = m; self._lastOutput = None                      # CutOff
        def f(m, i, o): self._lastOutput = o                                     # CutOff
        self.handle = self.m.register_forward_hook(f)                            # CutOff
    def forward(self, *args, **kwargs):                                          # CutOff
        self._lastOutput = None                                                  # CutOff
        self.net(*args, **kwargs)                                                # CutOff
        return self._lastOutput                                                  # CutOff
    def __del__(self): self.handle.remove()                                      # CutOff
@k1lib.patch(ModuleSelector)                                                     # CutOff
def cutOff(self) -> nn.Module:                                                   # cutOff
    """Creates a new network that returns the selected layer's output.
Example::

    xb = torch.randn(10, 2)
    m = nn.Sequential(nn.Linear(2, 5), nn.Linear(5, 4), nn.Linear(4, 6))
    m0 = m.select("#0").cutOff(); m1 = m.select("#1").cutOff()
    # returns (10, 6)
    m(xb).shape
    # returns (10, 5)
    m0(xb).shape == torch.Size([10, 5])
    # returns (10, 4)
    m1(xb).shape == torch.Size([10, 4])"""                                       # cutOff
    return CutOff(self.nn, self.modules("*") | cli.item() | cli.op().nn)         # cutOff



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import re, logging, warnings
from typing import Union, Tuple, List
from .atom import Atom, mo, NoFreeElectrons, OctetFull
from .substance import alkane, alcohol
import k1lib.cli as cli
__all__ = ["parse"]
mainNumsDict = {"meth": 1, "eth": 2, "prop": 3, "but": 4, "pent": 5, "hex": 6, "hept": 7, "oct": 8, "non": 9, "dec": 10, "undec": 11, "dodec": 12, "tridec": 13, "tetradec": 14, "pentadec": 15, "hexadec": 16, "heptadec": 17, "octadec": 18, "nonadec": 19}
bondsDict = {"ane": 1, "ene": 2, "yne": 3}
cis_trans = "(?P<cis_trans>[, \-]?((cis)|(trans))$)"
def join(a): return "(" + "|".join(f"({p})" for p in a) + ")"                    # join
pres1 = ["n-", "o-", "p-"]                                                       # join
pres2 = ["mono", "di", "tri", "tetra", "penta", "hexa"]                          # join
pres3 = ["iso", "cyclo", "sec-", "tert-"]                                        # join
pres = f"{join(pres1)}?[\-]?{join(pres2)}?[\-]?{join(pres3)}?"                   # join
main_pre = f"(?P<main_pre>(((cis-)|(trans-))?[0-9,\-])*{pres})"                  # join
#main_pre = "(?P<main_pre>(n-)?(cyclo))"                                         # join
mids = list(mainNumsDict.keys()) + ["benz", "naphthal", "pyridine", "xylenol", "ether", "amine"] # join
main_mid = "(?P<main_mid>" + "|".join(f"({m})" for m in mids) + ")"              # join
sufs = list(bondsDict.keys()) + ["anol", "ate", "adiene", "aldehyde", "anone"]   # join
main_suf = f"(?P<main_suf>{join(sufs)})"                                         # join
main = f"(?P<main>{main_pre}?{main_mid}{main_suf}?{cis_trans}?)$"; mainC = re.compile(f"^{main}") # join
branch_pre = f"(?P<branch_pre>(((cis-)|(trans-))?[0-9,\-])*{pres})"              # join
bMids = mids + ["fluoro", "chloro", "bromo", "iodo", "nitro", "phenyl", "perfluoro"] # join
branch_mid = "(?P<branch_mid>" + join(bMids) + ")"                               # join
branch_suf = "(?P<branch_suf>yl)"                                                # join
branch = f"(?P<branch>{branch_pre}?{branch_mid}{branch_suf}?[ ]?)"; branchC = re.compile(f"^{branch}$") # join
fullStrict = f"^{branch}*{main}"                                                 # join
fullStrictC = re.compile(fullStrict)                                             # join
full = f"{branch}*{main}"                                                        # join
fullC = re.compile(full)                                                         # join
def recognizeInit(ogS:str) -> Union[Tuple[List[str], List[str]], str]:           # recognizeInit
    """Tries to recognize the input sequence. Returns the sequence if can't understand, else
returns tuple (mainParts, branches)."""                                          # recognizeInit
    s = ogS; m = re.fullmatch(fullStrictC, s)                                    # recognizeInit
    if m is None: return ogS                                                     # recognizeInit
    s = s[:m.start("main")]                                                      # recognizeInit
    sMain = m.group("main")                                                      # recognizeInit
    sMains = [m.group("main_pre") or "", m.group("main_mid") or "", m.group("main_suf") or ""] # recognizeInit
    sMains = ["".join(sMains), *sMains]                                          # recognizeInit
                                                                                 # recognizeInit
    groups = []                                                                  # recognizeInit
    for _ in range(int(10)):                                                     # recognizeInit
        m = re.match(branch, s)                                                  # recognizeInit
        if m is None or m.end() == 0: break                                      # recognizeInit
        groups.append(s[:m.end()].strip("-,"))                                   # recognizeInit
        s = s[m.end():]                                                          # recognizeInit
    if len(s) != 0: return None                                                  # recognizeInit
    return [sMains, groups]                                                      # recognizeInit
def preLocs(sPre, *defaults) -> List[int]:                                       # preLocs
    """If ``sPre`` looks like "3,1,4-" or sth like that, then extract all the location infos.""" # preLocs
    answer = []                                                                  # preLocs
    try: answer = sPre.strip("-,").split(",") | cli.replace("-", "") | cli.toInt() | cli.deref() # preLocs
    except: pass                                                                 # preLocs
    return defaults if len(answer) == 0 else answer                              # preLocs
def detectModifiers(sPre:str, has:set) -> str:                                   # detectModifiers
    if "cyclo" in sPre: has.add("cyclo"); sPre = sPre.replace("cyclo", "")       # detectModifiers
    if "iso" in sPre: has.add("iso"); sPre = sPre.replace("iso", "")             # detectModifiers
    if "sec" in sPre: has.add("sec"); sPre = sPre.replace("sec", "")             # detectModifiers
    if "tert" in sPre: has.add("tert"); sPre = sPre.replace("tert", "")          # detectModifiers
    return sPre                                                                  # detectModifiers
def fixTridec(sPre:str, sMid:str) -> Tuple[str, str]:                            # fixTridec
    """Handles special case for tridec, tetradec, ..., as "tri" leaks into sPre and not sMid for detection""" # fixTridec
    if sMid == "dec": # special case for tridec, tetradec and whatnot            # fixTridec
        for prefix in ["tri", "tetra", "penta", "hexa"]:                         # fixTridec
            if sPre.endswith(prefix):                                            # fixTridec
                sPre = sPre[:-len(prefix)]                                       # fixTridec
                sMid = prefix + sMid; break                                      # fixTridec
    return sPre, sMid                                                            # fixTridec
def recognizeMain(a:Union[List[str], str]) -> Union[Atom, str]:                  # recognizeMain
    """Tries to recognize main part of the molecule (eg. n-cyclohexane). Returns
:class:`~k1lib.mo.Atom` if recognized, else the input string"""                  # recognizeMain
    s = ""                                                                       # recognizeMain
    try:                                                                         # recognizeMain
        if isinstance(a, str):                                                   # recognizeMain
            m = re.fullmatch(mainC, a)                                           # recognizeMain
            a = [a, m.group("main_pre"), m.group("main_mid"), m.group("main_suf")] # recognizeMain
        s, sPre, sMid, sSuf = a                                                  # recognizeMain
        sPre, sMid, sSuf = (sPre or "").strip().lstrip("-"), (sMid or "").strip(), (sSuf or "").strip() # recognizeMain
        has = set(); sPre = detectModifiers(sPre, has)                           # recognizeMain
        if sMid in mainNumsDict:                                                 # recognizeMain
            sPre, sMid = fixTridec(sPre, sMid)                                   # recognizeMain
            carbons = mainNumsDict[sMid]                                         # recognizeMain
            if sSuf in bondsDict:                                                # recognizeMain
                a = alkane(carbons); bonds = bondsDict[sSuf] # main atom         # recognizeMain
                if bonds > 1: b = a.next(times=preLocs(sPre, 1)[0]-1); b.bond(b.next(), bonds-1) # recognizeMain
            elif sSuf == "anol":                                                 # recognizeMain
                a = alcohol(carbons, preLocs(sPre, 1)[0]) # main atom            # recognizeMain
            elif sSuf == "adiene":                                               # recognizeMain
                a = alkane(carbons)                                              # recognizeMain
                for loc in preLocs(sPre, 1, 2):                                  # recognizeMain
                    b = a.next(times=loc-1); b(b.next())                         # recognizeMain
            else: return s                                                       # recognizeMain
        elif sMid == "xylenol" and sSuf.strip() == "":                           # recognizeMain
            a = mo.benzene; a(mo.H2O)                                            # recognizeMain
            for loc in preLocs(sPre): a.next(times=loc-1)(mo.CH4)                # recognizeMain
        else: return s                                                           # recognizeMain
        if "cyclo" in has: a.endChain.bond(a)                                    # recognizeMain
        if "iso" in has: a.moveLastCTo2ndC()                                     # recognizeMain
        if "tert" in has: a.moveLastCTo2ndC().moveLastCTo2ndC()                  # recognizeMain
        return a                                                                 # recognizeMain
    except Exception as e:                                                       # recognizeMain
        print(s); raise e                                                        # recognizeMain
diTriTetra = {term: i+1 for i, term in enumerate(pres2)}                         # recognizeMain
def detectDiTriTetra(sPre:str) -> Tuple[str, str]:                               # detectDiTriTetra
    """If exists, returns (sPre, num), else return (sPre, None)"""               # detectDiTriTetra
    for k, v in diTriTetra.items():                                              # detectDiTriTetra
        if sPre.endswith(k):                                                     # detectDiTriTetra
            return sPre.replace(k, ""), v                                        # detectDiTriTetra
    return sPre, None                                                            # detectDiTriTetra
def detectCisTrans(sPre:str, has) -> str:                                        # detectCisTrans
    for term in ["cis", "trans"]:                                                # detectCisTrans
        if term in sPre:                                                         # detectCisTrans
            has.add(term); sPre = sPre.replace(term, "")                         # detectCisTrans
    return sPre                                                                  # detectCisTrans
def recognizeGroup(s:str, mainA:Atom) -> Union[Atom, str]:                       # recognizeGroup
    """Tries to recognize a group (eg. "1-cyclopentyl"), and adds it to the main atom ``mainA``.

:return: if can't recognize, then returns the group string, else returns the main atom""" # recognizeGroup
    if mainA is None: mainA = alkane(10)                                         # recognizeGroup
    m = re.fullmatch(branchC, s); ops = [] # operations to generate the branch   # recognizeGroup
    sPre, sMid, sSuf = [m.group("branch_pre"), m.group("branch_mid"), m.group("branch_suf")] # recognizeGroup
    has = set(); sPre = detectCisTrans(detectModifiers(sPre, has), has)          # recognizeGroup
    if sMid in mainNumsDict:                                                     # recognizeGroup
        if sSuf != "yl": return s                                                # recognizeGroup
        carbons = mainNumsDict[sMid]                                             # recognizeGroup
        ops.append(lambda _: alkane(carbons))                                    # recognizeGroup
        if "cyclo" in has: ops.append(lambda a: a.endChain.bond(a))              # recognizeGroup
        if "iso" in has: ops.append(lambda a: a.moveLastCTo2ndC())               # recognizeGroup
        if "tert" in has: ops.append(lambda a: a.moveLastCTo2ndC().moveLastCTo2ndC()) # recognizeGroup
    elif sMid == "chloro": ops.append(lambda _: mo.Cl(mo.H))                     # recognizeGroup
    elif sMid == "fluoro": ops.append(lambda _: mo.F (mo.H))                     # recognizeGroup
    elif sMid == "bromo":  ops.append(lambda _: mo.Br(mo.H))                     # recognizeGroup
    elif sMid == "iodo":   ops.append(lambda _: mo.I (mo.H))                     # recognizeGroup
    elif sMid == "phenyl": ops.append(lambda _: mo.benzene)                      # recognizeGroup
    else: return s#, sPre, sMid, sSuf                                            # recognizeGroup
    def gen(a=None):                                                             # recognizeGroup
        for op in ops: a = op(a)                                                 # recognizeGroup
        return a                                                                 # recognizeGroup
    # now attaching group to main chain...                                       # recognizeGroup
    sPre, locWord = detectDiTriTetra(sPre); locs = preLocs(sPre)                 # recognizeGroup
    logging.debug(f"locs: {locs}, locWord: {locWord}, sPre: {sPre}")             # recognizeGroup
    if len(locs) > 0 and locWord is not None: # both available, check consistency # recognizeGroup
        logging.debug(f"locs and locWord both available")                        # recognizeGroup
        assert locWord == len(locs), f"s: {s}, locWord: {locWord}, locs: {locs}" # recognizeGroup
        for loc in locs: mainA.next(times=loc-1)(gen())                          # recognizeGroup
        return mainA                                                             # recognizeGroup
    elif len(locs) > 0 and locWord is None:                                      # recognizeGroup
        if len(locs) > 1: warnings.warn(f"Group {s} has attachment positions, but not di,tri,tetra words!") # recognizeGroup
        for loc in locs: mainA.next(times=loc-1)(gen())                          # recognizeGroup
        return mainA                                                             # recognizeGroup
    else:                                                                        # recognizeGroup
        logging.debug(f"Can't detect any position, will be adding stuff randomly to the molecule") # recognizeGroup
        a = mainA                                                                # recognizeGroup
        for _ in range(max(len(locs), locWord or 0, 1)):                         # recognizeGroup
            try:                                                                 # recognizeGroup
                while a.availableBonds == 0: a = a.next()                        # recognizeGroup
            except RecursionError: raise RuntimeError(f"Molecule has no more free carbons to attach group {s} to") # recognizeGroup
            a(gen())                                                             # recognizeGroup
            pass                                                                 # recognizeGroup
        return mainA                                                             # recognizeGroup
def parse(s:str, quiet:bool=False) -> Union[Atom, str]:                          # parse
    """Tries to recognize molecule. Returns molecule :class:`~k1lib.mo.Atom`
if recognized, else the molecule's string."""                                    # parse
    def p(s):                                                                    # parse
        if not quiet: print(s)                                                   # parse
    try:                                                                         # parse
        # initial recognition                                                    # parse
        out = recognizeInit(s); has = set()                                      # parse
        if isinstance(out, str): return s                                        # parse
        mainParts, groups = out                                                  # parse
        # form main chain                                                        # parse
        mainA = recognizeMain(mainParts)                                         # parse
        if isinstance(mainA, str):                                               # parse
            p(f"Recognized {s}, but couldn't form the main chain {mainParts[0]}"); return s # parse
        # check for special perfluoro branch case                                # parse
        gs = []                                                                  # parse
        for g in groups:                                                         # parse
            if "perfluoro" in g:                                                 # parse
                if "perfluoro" != g: p(f"For some reason, group {g} is not exactly 'perfluoro'."); return s # parse
                has.add("perfluoro")                                             # parse
            else: gs.append(g)                                                   # parse
        for g in gs:                                                             # parse
            mainA = recognizeGroup(g, mainA)                                     # parse
            if isinstance(mainA, str):                                           # parse
                p(f"Recognized {s}, the main chain too, but don't recognize group {g}") # parse
        if "perfluoro" in has: mainA.perfluoro_ize()                             # parse
        return mainA                                                             # parse
    except OctetFull:                                                            # parse
        p("Some of the atoms can't form new bonds, as their octet is full. Please revise your structure!"); return s # parse
    except NoFreeElectrons:                                                      # parse
        p("Some of the atoms can't form new bonds, as they have no free electrons left to contribute to the bond. Please revise your structure!"); return s # parse



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""All tools related to cif file format that describes protein structures.
Expected to use behind the "cif" module name, like this::

    from k1lib.imports import *
    cif.cat("abc.cif")
"""
import k1lib
import k1lib.cli as cli;
from k1lib.cli.init import BaseCli, yieldT
from typing import Iterator, List
from k1lib.cli.typehint import *
import k1lib._k1a as k1a
__all__ = ["tables", "records"]
hasTable = lambda: cli.filt(cli.grep("loop_") | cli.shape(0) | (cli.op() > 0))
toBlocks = lambda: cli.cat() | cli.grep("^#", sep=True).till()
def collect(l):                                                                  # collect
    inBlock = False; tmp = []                                                    # collect
    for e in l:                                                                  # collect
        if e.startswith(";"):                                                    # collect
            inBlock = not inBlock                                                # collect
            if not inBlock: yield ("".join(tmp))[1:]; continue                   # collect
        if not inBlock: yield e                                                  # collect
        else: tmp.append(e)                                                      # collect
def tables(name=None, dikt=True):                                                # tables
    """Loads table info.
Dictionary mode::

    # both return output below
    "1z7z.cif" | cif.tables() | op()["_audit_author"]
    "1z7z.cif" | cif.tables("_audit_author")

Potential output::

    {'name': ("'Xiao, C.'",
      "'Bator-Kelly, C.M.'",
      "'Rieder, E.'",
      "'Chipman, P.R.'",
      "'Craig, A.'",
      "'Kuhn, R.J.'",
      "'Wimmer, E.'",
      "'Rossmann, M.G.'"),
     'pdbx_ordinal': ('1', '2', '3', '4', '5', '6', '7', '8')}

Result is a dictionary of ``table name -> dict()``. That inner dictionary maps from
column name to a list of elements. All columns should have the same number of elements.

Table mode::

    # both return output below
    "1z7z.cif" | cif.tables("_audit_author", dikt=False)
    "1z7z.cif" | cif.tables(dikt=False) | op()["_audit_author"]

Potential output::

    [['name', 'pdbx_ordinal'],
     ["'Xiao, C.'", '1'],
     ["'Bator-Kelly, C.M.'", '2'],
     ["'Rieder, E.'", '3'],
     ["'Chipman, P.R.'", '4'],
     ["'Craig, A.'", '5'],
     ["'Kuhn, R.J.'", '6'],
     ["'Wimmer, E.'", '7'],
     ["'Rossmann, M.G.'", '8']]

Result is a dictionary of ``table name -> List[List[str]]``. So basically you're
getting the table directly.

:param name: if specified, only grabs the specified table, else returns every table
:param dikt: whether to return a dict or table for each table"""                 # tables
    def inner(url): # f is iden() or deref(), depending on perf characteristics that you want # tables
        a = url | toBlocks() | hasTable() | cli.apply(~cli.head(2) | cli.op().strip().all()) | cli.deref() # preprocessing, split to blocks # tables
        b = a | cli.apply(~cli.filt(cli.op().startswith("_")) | (cli.aS(k1a.str_split, " ") | cli.filt(cli.op() != "")).all() | cli.joinStreams()) # tables
        fieldss = a | cli.apply(cli.filt(cli.op().startswith("_")) | cli.op().split(".")[1].all()) | cli.toList().all() # tables
        f = (cli.transpose() | cli.apply(cli.item() & ~cli.head(1)) | cli.transpose() | cli.toDict(False)) if dikt else cli.iden() # asDict # tables
        tableNames = a | cli.op()[0].split(".")[0].all() | cli.deref()           # tables
        c = [b, fieldss, tableNames] | cli.transpose() | (cli.filt(cli.op() == name, 2) if name is not None else cli.iden()) | cli.cut(0, 1)\
            | ~cli.apply(lambda l, fields: collect(l) | cli.batched(len(fields), True) | cli.insert(fields) | f) # tables
        d = [tableNames, c] | cli.toDict(False)                                  # tables
        if name is None: return d                                                # tables
        else:                                                                    # tables
            if len(d) != 1: return None                                          # tables
            else: return d.values() | cli.item()                                 # tables
    return cli.aS(inner)                                                         # tables
def records():                                                                   # records
    """Load record info.
Example::

    "1z7z.cif" | cif.records() | op()["_exptl"] | deref()

Potential output::

    [['entry_id', '1Z7Z'],
     ['method', "'ELECTRON MICROSCOPY'"],
     ['crystals_number', '?']]

Result is a dictionary of ``record name -> (n, 2) table``"""                     # records
    each = ~cli.head(1) | cli.op().strip().all() | cli.apply(cli.aS(k1a.str_split, " ") | cli.filt(cli.op() != "")) | cli.joinStreams() | cli.aS(collect) | cli.batched(2)\
        | (cli.item(2) | cli.op().split(".")[0]) & (cli.apply(cli.op().split(".")[1], 0)) # records
    return toBlocks() | ~hasTable() | cli.filt(cli.op().ab_len() > 1) | each.all() | cli.toDict() # records



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This is for functions that cuts out specific parts of the table
"""
from typing import Callable, Union, List, overload, Iterator, Any, Set, Tuple
from k1lib.cli.init import BaseCli, fastF; import k1lib.cli.init as init
import k1lib.cli as cli; import k1lib, json, os, math, traceback
from k1lib.cli.typehint import *
import numpy as np; from collections import deque
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["filt", "filter_", "inSet", "contains", "empty",
           "isNumeric", "instanceOf",
           "head", "tail", "cut", "rows",
           "intersection", "union", "unique", "breakIf", "mask", "tryout", "resume",
           "trigger", "filtStd"]
settings = k1lib.settings.cli
class filt(BaseCli):                                                             # filt
    def __init__(self, predicate:Callable[[Any], bool], column:Union[int, List[int]]=None, catchErrors:bool=False): # filt
        """Filters out elements.
Examples::

    # returns [2, 6], grabbing all the even elements
    [2, 3, 5, 6] | filt(lambda x: x%2 == 0) | deref()
    # returns [3, 5], grabbing all the odd elements
    [2, 3, 5, 6] | ~filt(lambda x: x%2 == 0) | deref()
    # returns [[2, 'a'], [6, 'c']], grabbing all the even elements in the 1st column
    [[2, "a"], [3, "b"], [5, "a"], [6, "c"]] | filt(lambda x: x%2 == 0, 0) | deref()
    # throws error, because strings can't mod divide
    [1, 2, "b", 8] | filt(lambda x: x % 2 == 0) | deref()
    # returns [2, 8]
    [1, 2, "b", 8] | filt(lambda x: x % 2 == 0, catchErrors=True) | deref()

You can also pass in :class:`~k1lib.cli.modifier.op` or string, for extra intuitiveness and quickness::

    # returns [2, 6]
    [2, 3, 5, 6] | filt(op() % 2 == 0) | deref()
    # returns ['abc', 'a12']
    ["abc", "def", "a12"] | filt(op().startswith("a")) | deref()
    # returns [3, 4, 5, 6, 7, 8, 9]
    range(100) | filt(3 <= op() < 10) | deref()
    # returns [3, 4, 5, 6, 7, 8, 9]
    range(100) | filt("3 <= x < 10") | deref()

See :class:`~k1lib.cli.modifier.aS` for more details on string mode. If you
pass in :class:`numpy.ndarray` or :class:`torch.Tensor`, then it will
automatically use the C-accelerated versions if possible, like this::

    # returns np.array([2, 3, 4]), instead of iter([2, 3, 4])
    np.array([1, 2, 3, 4]) | filt(lambda x: x>=2) | deref()
    # returns [2, 3, 4], instead of np.array([2, 3, 4]), because `math.exp` can't operate on numpy arrays
    np.array([1, 2, 3, 4]) | filt(lambda x: math.exp(x) >= 3) | deref()

If you need more extensive filtering capabilities involving text, check out :class:`~k1lib.cli.grep.grep`

If "filt" is too hard to remember, this cli also has an alias :class:`filter_`
that kinda mimics Python's ``filter()``.

:param predicate: function that returns True or False
:param column: if not specified, then filters elements of the input
    array, else filters the specific column only (or columns, just
    like in :class:`~k1lib.cli.modifier.apply`)
:param catchErrors: whether to catch errors in the function or not (reject
    elements that raise errors). Runs slower if enabled though"""                # filt
        fs = [predicate]; super().__init__(fs); self.inverted = False; self.preInvFilt = None # filt
        if column:                                                               # filt
            ex = Exception(f"Filtering using a function on a negative-indexed column ({column}) is not supported") # filt
            if isinstance(column, int):                                          # filt
                if column < 0: raise ex                                          # filt
            else:                                                                # filt
                column = list(column)                                            # filt
                if len([c for c in column if c < 0]): raise ex                   # filt
        self.f = f = fs[0]; _fP = fastF(f); self.column = column                 # filt
        if catchErrors:                                                          # filt
            def g(x):                                                            # filt
                try: return _fP(x)                                               # filt
                except: return False                                             # filt
            self.predicate = g                                                   # filt
        else: self.predicate = _fP                                               # filt
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # filt
        p = self.predicate; c = self.column                                      # filt
        if c is None:                                                            # filt
            if isinstance(it, settings.arrayTypes):                              # filt
                try: return it[p(it)]                                            # filt
                except Exception as e: print(e)                                  # filt
            return (l for l in it if p(l))                                       # filt
        elif isinstance(c, int):                                                 # filt
            if isinstance(it, settings.arrayTypes):                              # filt
                try: return it[p(it[:,c])]                                       # filt
                except: pass                                                     # filt
            def gen():                                                           # filt
                for es in it:                                                    # filt
                    es = list(es)                                                # filt
                    if c < len(es) and p(es[c]): yield es                        # filt
            return gen()                                                         # filt
        else: # list of ints                                                     # filt
            ops = []                                                             # filt
            for c_ in c: ops.append(filt(self.predicate, c_, False))             # filt
            return it | cli.serial(*ops)                                         # filt
    def __invert__(self):                                                        # filt
        """Negate the condition"""                                               # filt
        if self.inverted: raise Exception("Can't invert filt() twice!")          # filt
        def f(s):                                                                # filt
            if isinstance(s, settings.arrayTypes):                               # filt
                res = self.predicate(s) # can cause an exception, but that's ok, as that's the signal telling the code in __ror__ to not pass in array types # filt
                if isinstance(res, settings.arrayTypes): return ~res             # filt
            return not self.predicate(s)                                         # filt
        ans = filt(f, self.column); ans.preInvFilt = self; ans.inverted = True; return ans # filt
    def __neg__(self):                                                           # filt
        """Also negates the condition"""                                         # filt
        return ~self                                                             # filt
    def split(self):                                                             # filt
        """Splits the input into positive and negative samples.
Example::

    # returns [[0, 2, 4, 6, 8], [1, 3, 5, 7, 9]]
    range(10) | filt(lambda x: x%2 == 0).split() | deref()
    # also returns [[0, 2, 4, 6, 8], [1, 3, 5, 7, 9]], exactly like above
    range(10) | filt(lambda x: x%2 == 0) & filt(lambda x: x%2 != 0) | deref()""" # filt
        f = self.predicate; c = self.column; return filt(f, c) & ~filt(f, c)     # filt
    def _jsF(self, meta):                                                        # filt
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); argIdx = init._jsDAuto(); inverted = False # filt
        if self.inverted: self = self.preInvFilt; inverted = True                # filt
        header, _fIdx, _async = k1lib.kast.asyncGuard(k1lib.kast.prepareFunc3(self.f, ("filt", meta))) # filt
        return f"{header}\nconst {fIdx} = {'async ' if _async else ''}({dataIdx}) => {'await ' if _async else ''}{dataIdx}.filt{'_async' if _async else ''}({'async ' if _async else ''}({argIdx}) => {'!' if inverted else ''}({'await ' if _async else ''}{_fIdx}({argIdx})), {cli.kjs.v(self.column)})", fIdx # filt
filter_ = filt                                                                   # filt
class inSet(filt):                                                               # inSet
    def __init__(self, values:Set[Any], column:int=None, inverse=False):         # inSet
        """Filters out lines that is not in the specified set.
Example::

    # returns [2, 3]
    range(5) | inSet([2, 8, 3]) | deref()
    # returns [0, 1, 4]
    range(5) | ~inSet([2, 8, 3]) | deref()"""                                    # inSet
        if not isinstance(values, (set, dict)): values = set(values)             # inSet
        super().__init__((lambda l: l in values) if not inverse else (lambda l: not l in values), column) # inSet
        self.values = values; self.column = column; self.inverse = inverse       # inSet
    def __invert__(self): return inSet(self.values, self.column, not self.inverse) # inSet
    def _jsF(self, meta):                                                        # inSet
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); setIdx = init._jsDAuto() # inSet
        return f"const {setIdx} = {json.dumps(list(self.values))};\nconst {fIdx} = ({dataIdx}) => {dataIdx}.inSet({setIdx}, {cli.kjs.v(self.column)}, {cli.kjs.v(self.inverse)})", fIdx # inSet
class contains(filt):                                                            # contains
    def __init__(self, s:str, column:int=None, inverse=False):                   # contains
        """Filters out lines that don't contain the specified substring. Sort of similar
to :class:`~k1lib.cli.grep.grep`, but this is simpler, and can be inverted.
Example::

    # returns ['abcd', '2bcr']
    ["abcd", "0123", "2bcr"] | contains("bc") | deref()"""                       # contains
        super().__init__((lambda e: s in e) if not inverse else (lambda e: not s in e), column) # contains
        self.s = s; self.column = column; self.inverse = inverse                 # contains
    def __invert__(self): return contains(self.s, self.column, not self.inverse) # contains
    def _jsF(self, meta):                                                        # contains
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); sIdx = init._jsDAuto() # contains
        return f"const {sIdx} = {json.dumps(self.s)};\nconst {fIdx} = ({dataIdx}) => {dataIdx}.contains({sIdx}, {cli.kjs.v(self.column)}, {cli.kjs.v(self.inverse)})", fIdx # contains
class empty(BaseCli):                                                            # empty
    def __init__(self, reverse=False):                                           # empty
        """Filters out streams that is not empty. Almost always used inverted,
but "empty" is a short, sweet name that's easy to remember. Example::

    # returns [[1, 2], ['a']]
    [[], [1, 2], [], ["a"]] | ~empty() | deref()

:param reverse: not intended to be used by the end user. Do ``~empty()`` instead.""" # empty
        super().__init__(); self.reverse = reverse                               # empty
    def __ror__(self, streams:Iterator[Iterator[Any]]) -> Iterator[Iterator[Any]]: # empty
        r = self.reverse                                                         # empty
        for stream in streams:                                                   # empty
            try:                                                                 # empty
                item, it = stream | cli.peek()                                   # empty
                if not r:                                                        # empty
                    if it == []: yield it                                        # empty
                else:                                                            # empty
                    if it != []: yield it                                        # empty
            except StopIteration: pass                                           # empty
    def __invert__(self):                                                        # empty
        return empty(not self.reverse)                                           # empty
def isNumeric(column:int=None) -> filt:                                          # isNumeric
    """Filters out a line if that column is not a number.
Example::

    # returns [0, 2, '3']
    [0, 2, "3", "a"] | isNumeric() | deref()"""                                  # isNumeric
    def f(v):                                                                    # isNumeric
        try: float(v); return True                                               # isNumeric
        except ValueError: return False                                          # isNumeric
    return filt(f, column)                                                       # isNumeric
def instanceOf(cls:Union[type, Tuple[type]], column:int=None) -> filt:           # instanceOf
    """Filters out lines that is not an instance of the given type.
Example::

    # returns [2]
    [2, 2.3, "a"] | instanceOf(int) | deref()
    # returns [2, 2.3]
    [2, 2.3, "a"] | instanceOf((int, float)) | deref()"""                        # instanceOf
    if isinstance(cls, list): cls = tuple(cls)                                   # instanceOf
    return filt(lambda e: isinstance(e, cls), column)                            # instanceOf
def sliceable(it):                                                               # sliceable
    try: it[:]; len(it); return True                                             # sliceable
    except: return False                                                         # sliceable
def _head(n, inverted, it):                                                      # _head
    it = iter(it)                                                                # _head
    if n is None:                                                                # _head
        if not inverted: yield from it                                           # _head
        else: return                                                             # _head
    elif n >= 0:                                                                 # _head
        if not inverted:                                                         # _head
            for i, line in zip(range(n), it): yield line                         # _head
        else:                                                                    # _head
            for i, line in zip(range(n), it): pass                               # _head
            yield from it                                                        # _head
    else:                                                                        # _head
        if not inverted: # head to -3                                            # _head
            n = abs(n); queue = deque()                                          # _head
            for line in it:                                                      # _head
                queue.append(line)                                               # _head
                if len(queue) > n: yield queue.popleft()                         # _head
        else: yield from deque(it, -n) # -3 to end                               # _head
class head(BaseCli):                                                             # head
    def __init__(self, n=10):                                                    # head
        """Only outputs first ``n`` elements. You can also negate it (like
``~head(5)``), which then only outputs after first ``n`` lines. Examples::

    "abcde" | head(2) | deref() # returns ["a", "b"]
    "abcde" | ~head(2) | deref() # returns ["c", "d", "e"]
    "0123456" | head(-3) | deref() # returns ['0', '1', '2', '3']
    "0123456" | ~head(-3) | deref() # returns ['4', '5', '6']
    "012" | head(None) | deref() # returns ['0', '1', '2']
    "012" | ~head(None) | deref() # returns []

You can also pass in fractional head::

    range(20) | head(0.25) | deref() # returns [0, 1, 2, 3, 4], or the first 25% of samples

Also works well and fast with :class:`numpy.ndarray`, :class:`torch.Tensor`
and other sliceable types::

    # returns (10,)
    np.linspace(1, 3) | head(10) | shape()"""                                    # head
        super().__init__(); self.n = n; self.inverted = False; self._sliceable = None # head
    def _all_array_opt(self, it, level):                                         # head
        n = self.n; inverted = self.inverted                                     # head
        if n is not None and round(n) != n: n = int(it.shape[level]*n) # fractional head # head
        sl = tuple([slice(None)]*level); return it[(*sl, slice(n, None))] if inverted else it[(*sl, slice(None, n))] # head
    def _typehint(self, inp):                                                    # head
        if isinstance(inp, tListIter): return inp                                # head
        if isinstance(inp, tArrayTypes): return inp                              # head
        if inp == str: return str                                                # head
        return tIter(tAny())                                                     # head
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # head
        n = self.n; inverted = self.inverted                                     # head
        if n is not None and round(n) != n: # fractional head                    # head
            if not sliceable(it): it = list(it)                                  # head
            i = int(len(it)*n)                                                   # head
            return it[i:] if inverted else it[:i]                                # head
        if inverted and n is None: return [] # special case                      # head
        _sliceable = self._sliceable # all of this to cache sliceable value, because it takes a lot of time to determine whether it's sliceable or not # head
        if _sliceable is None: self._sliceable = _sliceable = sliceable(it)      # head
        if _sliceable: return it[n:] if inverted else it[:n]                     # head
        else: return _head(self.n, self.inverted, it)                            # head
    def __invert__(self):                                                        # head
        h = head(self.n); h.inverted = not self.inverted                         # head
        return h                                                                 # head
    def split(self):                                                             # head
        """Splits the list up into a head and tail sections.
Example::

    # returns [[0, 1, 2, 3], [4, 5, 6, 7, 8, 9]]
    range(10) | head(4).split() | deref()

This only splits it into 2 parts. If you want to split it up
into many more parts with specified checkpoints, check out
:class:`~k1lib.cli.structural.splitC`."""                                        # head
        # return self & ~self # old version                                      # head
        return headSplit(self.n, self.inverted)                                  # head
    def _jsF(self, meta):                                                        # head
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); argIdx = init._jsDAuto() # head
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.head({self.n}, {cli.kjs.v(self.inverted)})", fIdx # head
class headSplit(BaseCli):                                                        # headSplit
    def __init__(self, n, inverted):                                             # headSplit
        self.n = n; self.inverted = inverted                                     # headSplit
        self.fixup = n is None or isinstance(n, float) or n < 0                  # headSplit
        self.sliceable = None                                                    # headSplit
    def _all_array_opt(self, it, level):                                         # headSplit
        n = self.n; inverted = self.inverted                                     # headSplit
        if n is not None and round(n) != n: n = int(it.shape[level]*n) # fractional head # headSplit
        sl = tuple([slice(None)]*level); b = it[(*sl, slice(n, None))]; a = it[(*sl, slice(None, n))] # headSplit
        return [b, a] if inverted else [a, b]                                    # headSplit
    def __ror__(self, it):                                                       # headSplit
        sliceable_ = self.sliceable; n = self.n                                  # headSplit
        if sliceable_ is None: self.sliceable = sliceable_ = sliceable(it)       # headSplit
        it = it if sliceable_ else list(it)                                      # headSplit
        if self.fixup: # needs to fix n to a more definite value. Just to make it faster # headSplit
            l = len(it)                                                          # headSplit
            if n is None: return it, []                                          # headSplit
            if isinstance(n, float): n = int(l*n) # fractional head              # headSplit
            n = (n+l)%l                                                          # headSplit
        return it[:n], it[n:]                                                    # headSplit
class tail(BaseCli):                                                             # tail
    def __init__(self, n:int=10):                                                # tail
        """Basically an inverted :class:`head`.
Examples::

    range(10) | tail(3) | deref() # returns [7, 8, 9]"""                         # tail
        self.n = n                                                               # tail
    def __ror__(self, it): return it | ~head(-self.n)                            # tail
    def _jsF(self, meta):                                                        # tail
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); argIdx = init._jsDAuto() # tail
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.head(-({self.n}), true)", fIdx # tail
class lazyList:                                                                  # lazyList
    def __init__(self, it):                                                      # lazyList
        self.it = iter(it); self.elems = []                                      # lazyList
    def __getitem__(self, idx):                                                  # lazyList
        elems = self.elems; it = self.it                                         # lazyList
        for _ in range(len(elems)-1, idx): elems.append(next(it))                # lazyList
        return elems[idx]                                                        # lazyList
class rows(BaseCli):                                                             # rows
    def __init__(self, *rows:List[int]):                                         # rows
        """Selects specific elements given an iterator of indexes.
Space complexity O(1) as a list is not constructed (unless you're
slicing it in really weird way). Example::

    "0123456789" | rows(2) | toList() # returns ["2"]
    "0123456789" | rows(5, 8) | toList() # returns ["5", "8"]
    "0123456789" | rows()[2:5] | toList() # returns ["2", "3", "4"]
    "0123456789" | ~rows()[2:5] | toList() # returns ["0", "1", "5", "6", "7", "8", "9"]
    "0123456789" | ~rows()[:7:2] | toList() # returns ['1', '3', '5', '7', '8', '9']
    "0123456789" | rows()[:-4] | toList() # returns ['0', '1', '2', '3', '4', '5']
    "0123456789" | ~rows()[:-4] | toList() # returns ['6', '7', '8', '9']

Why it's called "rows" is because I couldn't find a good name for
it. There was :class:`cut`, which the name of an actual bash cli
that selects out columns given indicies. When I needed a way to
do what this cli does, it was in the context of selecting out rows,
so the name stuck.

If you want to just pick out the nth item from the iterator, instead of doing
this::

    iter(range(10)) | rows(3) | item() # returns 3

... you can use the shorthand :class:`~k1lib.cli.utils.rItem` instead::

    iter(range(10)) | rItem(3) # returns 3

:param rows: ints for the row indices"""                                         # rows
        if len(rows) == 1 and isinstance(rows[0], slice):                        # rows
            self.slice = rows[0]; self.rows = None; self.idxMode = False         # rows
        else: self.slice = None; self.rows = rows; self.sortedRows = sorted(rows); self.idxMode = True # rows
        self.inverted = False                                                    # rows
    def __getitem__(self, _slice):                                               # rows
        start, stop, step = _slice.start, _slice.stop, _slice.step               # rows
        if step == None or step == 1:                                            # rows
            if start == None and stop == None: return cli.iden()                 # rows
            if start == None: return head(stop)                                  # rows
            if stop == None: return ~head(start)                                 # rows
        elif step == 0: return cli.ignore()                                      # rows
        answer = rows(_slice); answer.inverted = self.inverted; return answer    # rows
    def _all_array_opt(self, it, level:int):                                     # rows
        a = np.array(self.rows) if self.rows else self.slice; s = [slice(None, None, None)]*level # rows
        if self.inverted: mask = np.ones(it.shape[level], dtype=bool); mask[a] = False; return it[(*s, mask)] # rows
        return it[(*s, a)]                                                       # rows
    def __invert__(self): self.inverted = not self.inverted; return self         # rows
    def __ror__(self, it:Iterator[str]):                                         # rows
        idxMode = self.idxMode; inverted = self.inverted; sl = self.slice; rw = self.rows # rows
        def gen(it):                                                             # rows
            if not inverted:                                                     # rows
                if idxMode:                                                      # rows
                    if len(self.sortedRows) == 0: return                         # rows
                    it = list(it) if self.sortedRows[0] < 0 else lazyList(it)    # rows
                    for idx in rw: yield it[idx]                                 # rows
                else: yield from list(it)[sl]                                    # rows
            else:                                                                # rows
                it = list(it); n = len(it); idxs = set((e if e >= 0 else n+e) for e in rw) if idxMode else set(range(n)[sl]) # rows
                yield from (e for i, e in enumerate(it) if i not in idxs)        # rows
        if isinstance(it, settings.arrayTypes):                                  # rows
            a = np.array(rw) if rw else sl                                       # rows
            if inverted: mask = np.ones(len(it), dtype=bool); mask[a] = False; return it[mask] # rows
            else: return it[a]                                                   # rows
        return gen(it)                                                           # rows
    def _jsF(self, meta):                                                        # rows
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # rows
        if not self.slice is None: raise Exception("rows._jsF() doesn't support slice-based indexing yet") # rows
        if self.inverted: raise Exception("rows._jsF() doesn't support inversion yet") # rows
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.rows({cli.kjs.vs(self.rows) | cli.join(', ')})", fIdx # rows
class cut(BaseCli):                                                              # cut
    def __init__(self, *columns:List[int]):                                      # cut
        """Cuts out specific columns, sliceable. Examples::

    ["0123456789", "abcdefghij"] | cut(5, 8) | deref() # returns [['5', '8'], ['f', 'i']]
    ["0123456789", "abcdefghij"] | cut(8, 5) | deref() # returns [['8', '5'], ['i', 'f']], demonstrating permutation-safe
    ["0123456789"] | cut(5, 8) | deref() # returns [['5', '8']]
    ["0123456789"] | cut(8, 5) | deref() # returns [['8', '5']], demonstrating permutation-safe
    ["0123456789", "abcdefghij"] | cut(2) | deref() # returns ['2', 'c'], instead of [['2'], ['c']] as usual
    ["0123456789"] | cut(2) | deref() # returns ['2']
    ["0123456789"] | cut(5, 8) | deref() # returns [['5', '8']]
    ["0123456789"] | ~cut()[:7:2] | deref() # returns [['1', '3', '5', '7', '8', '9']]

In the first example, you can imagine that we're operating on this table::

    0123456789
    abcdefghij

Then, we want to grab the 5th and 8th column (0-indexed), which forms this table::

    58
    fi

So, result of that is just ``[['5', '8'], ['f', 'i']]``

In the fourth example, if you're only cutting out 1 column, then it
will just grab that column directly, instead of putting it in a list.

If you pass in :class:`numpy.ndarray` or :class:`torch.Tensor`, then it will
automatically use the C-accelerated versions, like this::

    torch.randn(4, 5, 6) | cut(2, 3)  # returns tensor of shape (4, 2, 6)
    torch.randn(4, 5, 6) | cut(2)     # returns tensor of shape (4, 6)
    torch.randn(4, 5, 6) | ~cut()[2:] # returns tensor of shape (4, 2, 6)

.. warning::

    TD;DR: inverted negative indexes are a bad thing when rows don't have the same number of elements

    Everything works fine when all of your rows have the same number of elements. But things might behave a
    little strangely if they don't. For example::

        # returns [['2', '3', '4'], ['2', '3', '4', '5', '6', '7']]. Different number of columns, works just fine
        ["0123456", "0123456789"]    |  cut()[2:-2] | deref()
        # returns [['0', '1', '8', '9'], ['a', 'b', 'i', 'j']]. Same number of columns, works just fine
        ["0123456789", "abcdefghij"] | ~cut()[2:-2] | deref()
        # returns [['0', '1', '5', '6'], ['0', '1', '5', '6', '7', '8', '9']]. Different number of columns, unsupported invert case
        ["0123456", "0123456789"]    | ~cut()[2:-2] | deref()

    Why does this happen? It peeks at the first row, determines that ~[2:-2] is equivalent
    to [:2] and [5:] combined and not [:2] and [-2:] combined. When applied to the second row,
    [-2:] goes from 5->9, hence the result. Another edge case would be::

        # returns [['0', '1', '2', '3', '5', '6'], ['0', '1', '2', '3', '5', '6', '7', '8', '9']]
        ["0123456", "0123456789"] | ~cut(-3) | deref()

    Like before, it peeks the first row and translate ~(-3) into ~4, which is equivalent to [:4] and [5:].
    But when applied to the second row, it now carries the meaning ~4, instead of ~(-3).

    Why don't I just fix these edge cases? Because the run time for it would be completely unacceptable,
    as we'd have to figure out what's the columns to include in the result for every row. This could
    easily be O(n^3). Of course, with more time optimizing, this could be solved, but this is the only
    extreme edge case and I don't feel like putting in the effort to optimize it.""" # cut
        super().__init__()                                                       # cut
        if len(columns) == 1 and isinstance(columns[0], slice): columns = columns[0] # columns is either a slice object, or a list of ints # cut
        self.columns = columns; self.inverted = False # columns: list[int] | slice # cut
    def _all_array_opt(self, it, level):                                         # cut
        c = self.columns; r = rows(c) if isinstance(c, slice) else rows(*c)      # cut
        if self.inverted: r = ~r                                                 # cut
        it = it | r.all(level+1); return (it | cli.item().all(level+1)) if not isinstance(c, slice) and len(c) == 1 else it # cut
    def __ror__(self, it):                                                       # cut
        columns = self.columns; inverted = self.inverted                         # cut
        isArray = isinstance(it, settings.arrayTypes)#; isArray = False          # cut
        if isArray: nCols = len(it[0]); prs = rs = range(nCols) # range(nColumns). "prs" for padded rs # cut
        else: # carefully peaking first row and get the number of columns        # cut
            it = iter(it); sentinel = object(); row = next(it, sentinel)         # cut
            if row is sentinel: return []                                        # cut
            row = list(row); it = it | cli.insert(row); nCols = len(row)         # cut
            rs = range(nCols); prs = range(nCols+20) # 20 for longer rows below. Also "rs" is not a great name, deal with it # cut
        if isinstance(columns, slice):                                           # cut
            if not inverted: return it[:,columns] if isArray else (row[columns] for row in (list(row) for row in it)) # cut
            columns = list(set(rs[columns]))                                     # cut
        columns = [e if e >= 0 else nCols + e for e in columns] # clear negative indicies # cut
        if self.inverted: columns = list(set(e for e in prs if e not in columns)) # cut
        if len(columns) == 1: c = columns[0]; return it[:,c] if isArray else (r[c] for r in (list(row) for row in it) if len(r) > c) # cut
        else: return it[:,columns] if isArray else ([row[c] for c in columns if c < len(row)] for row in (list(row) for row in it)) # cut
    def __getitem__(self, idx): answer = cut(idx); answer.inverted = self.inverted; return answer # cut
    def __invert__(self): self.inverted = not self.inverted; return self         # cut
    def _jsF(self, meta):                                                        # cut
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # cut
        if isinstance(self.columns, slice): raise Exception("cut._jsF() doesn't support slice-based indexing yet") # cut
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.{'cutInv' if self.inverted else 'cut'}({cli.kjs.vs(self.columns) | cli.join(', ')})", fIdx # cut
class intersection(BaseCli):                                                     # intersection
    def __init__(self, column=None, full=False):                                 # intersection
        """Returns the intersection of multiple streams.
Example::

    # returns set([2, 4, 5])
    [[1, 2, 3, 4, 5], [7, 2, 4, 6, 5]] | intersection()
    # returns ['2g', '4h', '5j']
    [["1a", "2b", "3c", "4d", "5e"], ["7f", "2g", "4h", "6i", "5j"]] | intersection(0) | deref()

If you want the full distribution, meaning the intersection, as well
as what's left of each stream, you can do something like this::

    # returns [{2, 4, 5}, [1, 3], [7, 6]]
    [[1, 2, 3, 4, 5], [7, 2, 4, 6, 5]] | intersection(full=True) | deref()

:param column: what column to apply the intersection
    on. Defaulted to None
:param full: if specified, return the full distribution, instead of the intersection alone""" # intersection
        super().__init__(); self.column = column                                 # intersection
        self.full = full                                                         # intersection
        self.f = intersection(column, False) if full else None                   # intersection
    def _typehint(self, inp):                                                    # intersection
        if self.column is None:                                                  # intersection
            if isinstance(inp, tArrayTypes): return tSet(inp.child)              # intersection
            if isinstance(inp, tListIterSet):                                    # intersection
                if isinstance(inp.child, tListIterSet):                          # intersection
                    return tSet(inp.child.child)                                 # intersection
                return tSet(tAny())                                              # intersection
            if isinstance(inp, tCollection):                                     # intersection
                a = inp.children[0]                                              # intersection
                for e in inp.children:                                           # intersection
                    if not isinstance(e, tListIterSet): return tSet(tAny())      # intersection
                    if e.child != a.child: return tSet(tAny())                   # intersection
                return tSet(a.child)                                             # intersection
            return tSet(tAny());                                                 # intersection
        else: return tAny()                                                      # intersection
    def __ror__(self, its:Iterator[Iterator[Any]]) -> Set[Any]:                  # intersection
        c = self.column; full = self.full; f = self.f                            # intersection
        if full:                                                                 # intersection
            if c is None:                                                        # intersection
                its = its | cli.deref(2); inter = its | f                        # intersection
                return [inter, *((e for e in it if e not in inter) for it in its)] # intersection
            else: raise Exception("intersection(int, True) mode not supported yet, as it's a little ambiguous what's the use case is, and there're many styles of functionality that this can take on") # intersection
        if c is None:                                                            # intersection
            answer = None                                                        # intersection
            for it in its:                                                       # intersection
                if answer is None: answer = set(it); continue                    # intersection
                answer = answer.intersection(it)                                 # intersection
            return set() if answer is None else answer                           # intersection
        else:                                                                    # intersection
            its = its | cli.deref(2); ans = {}                                   # intersection
            ids = its | cut(c).all() | intersection() | cli.aS(set)              # intersection
            for it in its:                                                       # intersection
                for row in it:                                                   # intersection
                    e = row[c]                                                   # intersection
                    if e in ids: ans[e] = row                                    # intersection
            return ans.values()                                                  # intersection
    def _jsF(self, meta):                                                        # intersection
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # intersection
        if not self.column is None: raise Exception("intersection._jsF() doesn't support targeting specific .column yet") # intersection
        if not self.full: raise Exception("intersection._jsF() doesn't support .full yet") # intersection
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.intersection()", fIdx   # intersection
class union(BaseCli):                                                            # union
    def __init__(self):                                                          # union
        """Returns the union of multiple streams.
Example::

    # returns {0, 1, 2, 10, 11, 12, 13, 14}
    [range(3), range(10, 15)] | union()
"""                                                                              # union
        super().__init__()                                                       # union
    def _typehint(self, inp):                                                    # union
        return intersection()._typehint(inp)                                     # union
    def __ror__(self, its:Iterator[Iterator[Any]]) -> Set[Any]:                  # union
        answer = set()                                                           # union
        for it in its: answer = set.union(answer, set(it))                       # union
        return answer                                                            # union
    def _jsF(self, meta):                                                        # union
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # union
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.union()", fIdx          # union
class unique(BaseCli):                                                           # unique
    def __init__(self, column:int=None):                                         # unique
        """Filters out non-unique row elements.
Example::

    # returns [[1, "a"], [2, "a"]]
    [[1, "a"], [2, "a"], [1, "b"]] | unique(0) | deref()
    # returns [0, 1, 2, 3, 4]
    [*range(5), *range(3)] | unique() | deref()

In the first example, because the 3rd element's first column is
1, which has already appeared, so it will be filtered out.

:param column: the column to detect unique elements. Can be
    None, which will behave like converting the input iterator
    into a set, but this cli will maintain the order"""                          # unique
        super().__init__(); self.column = column                                 # unique
    def __ror__(self, it):                                                       # unique
        c = self.column                                                          # unique
        if c is None:                                                            # unique
            if isinstance(it, settings.arrayTypes): bm = np if isinstance(it, np.ndarray) else (torch if hasTorch and isinstance(it, torch.Tensor) else None); return bm.unique(it) # unique
            def gen():                                                           # unique
                terms = set()                                                    # unique
                for e in it:                                                     # unique
                    if e not in terms: yield e                                   # unique
                    terms.add(e)                                                 # unique
        else:                                                                    # unique
            def gen():                                                           # unique
                terms = set()                                                    # unique
                for row in it:                                                   # unique
                    row = list(row); e = row[c]                                  # unique
                    if e not in terms: yield row                                 # unique
                    terms.add(e)                                                 # unique
        return gen()                                                             # unique
    def _jsF(self, meta):                                                        # unique
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # unique
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.unique({cli.kjs.v(self.column)})", fIdx # unique
class breakIf(BaseCli):                                                          # breakIf
    def __init__(self, f, col:int=None):                                         # breakIf
        """Breaks the input iterator if a condition is met.
Example::

    # returns [0, 1, 2, 3, 4, 5]
    [*range(10), 2, 3] | breakIf(lambda x: x > 5) | deref()
    # returns [[1, 'a'], [2, 'b']]
    [[1, "a"], [2, "b"], [3, "c"], [2, "d"], [1, "e"]] | breakIf("x > 2", 0) | deref()

:param col: column to apply the condition on"""                                  # breakIf
        fs = [f]; super().__init__(fs); self.f = fs[0]; self._fC = fastF(self.f); self.col = col # breakIf
    def _typehint(self, inp):                                                    # breakIf
        if isinstance(inp, tListIterSet): return tIter(inp.child)                # breakIf
        return tIter(tAny())                                                     # breakIf
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # breakIf
        f = self._fC; col = self.col                                             # breakIf
        if col is None:                                                          # breakIf
            for line in it:                                                      # breakIf
                if f(line): break                                                # breakIf
                yield line                                                       # breakIf
        else:                                                                    # breakIf
            for row in it:                                                       # breakIf
                if f(row[col]): break                                            # breakIf
                yield row                                                        # breakIf
    def _jsF(self, meta):                                                        # breakIf
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); argIdx = init._jsDAuto() # breakIf
        header, _fIdx = k1lib.kast.prepareFunc3(self.f, ("breakIf", meta))       # breakIf
        return f"{header}\nconst {fIdx} = ({dataIdx}) => {dataIdx}.breakIf(({argIdx}) => {_fIdx}({argIdx}))", fIdx # breakIf
class mask(BaseCli):                                                             # mask
    def __init__(self, mask:Iterator[bool]):                                     # mask
        """Masks the input stream.
Example::

    # returns [0, 1, 3]
    range(5) | mask([True, True, False, True, False])  | deref()
    # returns [2, 4]
    range(5) | ~mask([True, True, False, True, False]) | deref()
    # returns torch.tensor([0, 1, 3])
    torch.tensor(range(5)) | mask([True, True, False, True, False])"""           # mask
        super().__init__(); self.mask = mask; self.inverted = False              # mask
    def __invert__(self): res = mask(self.mask); res.inverted = not self.inverted; return res # mask
    def __ror__(self, it):                                                       # mask
        if self.inverted:                                                        # mask
            if isinstance(it, settings.arrayTypes): return it[[not e for e in self.mask]] # mask
            return (e for e, m in zip(it, self.mask) if not m)                   # mask
        else:                                                                    # mask
            if isinstance(it, settings.arrayTypes): return it[list(self.mask)]   # mask
            return (e for e, m in zip(it, self.mask) if m)                       # mask
class tryout(BaseCli):                                                           # tryout
    def __init__(self, result=None, retries=0, mode="result"):                   # tryout
        """Wraps every cli operation after this in a try-catch block, returning ``result``
if the operation fails. Example::

    # returns 9
    3 | (tryout("failed") | op()**2)
    # returns "failed", instead of raising an exception
    "3" | (tryout("failed") | op()**2)
    # special mode: returns "unsupported operand type(s) for ** or pow(): 'str' and 'int'"
    "3" | (tryout(mode="str") | op()**2)
    # special mode: returns entire trace stack (do `import traceback` first)
    "3" | (tryout(mode="traceback") | op()**2)
    # special mode: returns "3", the input of the tryout() block
    "3" | (tryout(mode="input") | op()**2)

By default, this ``tryout()`` object will gobble up all clis behind it and wrap
them inside a try-catch block. This might be undesirable, so you can stop it early::

    # returns "failed"
    3 | (tryout("failed") | op()**2 | aS(str) | op()**2)
    # raises an exception, because it errors out after the tryout()-captured operations
    3 | (tryout("failed") | op()**2) | aS(str) | op()**2

In the first example, :class:`tryout` will catch any errors happening within ``op()``,
``aS(str)`` or the second ``op()**2``. In the second example, :class:`tryout` will only
catch errors happening within the first ``op()**2``.

.. admonition:: Array mode

    The above works well for atomic operations and not looping operations. Let's
    say we have this function::

        counter = 0
        def f(x):
            global counter
            if x > 5:
                counter += 1
                if counter < 3: raise Exception(f"random error: {x}")
            return x**2

    This code will throw an error if x is greater than 5 for the first and second
    time (but runs smoothly after that. It's a really nasty function I know).
    Capturing like this will work::

        counter = 0 # line below returns [0, 1, 4, 9, 16, 25, 'failed', 'failed', 64, 81]
        range(10) | apply(tryout("failed") | aS(f)) | deref()

    But capturing like this won't work::

        counter = 0 # line below throws an exception
        range(10) | (tryout("failed") | apply(f)) | deref()

    The reason being, :class:`tryout` will only capture errors when the data is passed
    into ``apply(f)``, and won't capture it later on. However, when data is passed to
    ``apply(f)``, it hasn't executed anything yet (remember these things are lazily
    executed). So the exception actually happens when you're trying to ``deref()`` it,
    which lies outside of :class:`tryout`'s reach. You can just put a tilde in front
    to tell it to capture errors for individual elements in the iterator::

        counter = 0 # line belows returns [0, 1, 4, 9, 16, 25, 'failed', 'failed', 64, 81]
        range(10) | (~tryout("failed") | apply(f)) | deref()

    This mode has a weird quirk that requires that there has to be a 1-to-1 correspondence
    between the input and output for the block of code that it wraps around. Meaning this is okay::

        def g(x):
            global counter
            if 40 > x[0] >= 30:
                counter += 1
                if counter < 5: raise Exception("random error")
            return x
        counter = 0 # returns 50, corrects errors as if it's not even there!
        range(50) | (~tryout(None, 6) | batched(10, True) | apply(g) | joinStreams()) | deref() | shape(0)

    This is okay because going in, there're 50 elements, and it's expected that 50 elements
    goes out of :class:`tryout`. The input can be of infinite length, but there has to be a
    1-to-1 relationship between the input and output. While this is not okay::

        counter = 0 # returns 75, data structure corrupted
        range(50) | (~tryout(None, 6) | batched(10, True) | apply(g) | joinStreams() | batched(2, True)) | joinStreams() | deref() | shape(0)

    It's not okay because it's expected that 25 pairs of elements goes out of :class:`tryout`

.. admonition:: Retrying

    There's also the ``retries`` parameter, which specifies how many times should this
    class retry the operation until actually returning the predefined result::

        counter = 0 # line below returns [0, 1, 4, 9, 16, 25, None, None, 64, 81]
        range(10) | (~tryout(retries=0) | apply(f)) | deref()
        counter = 0 # line below returns [0, 1, 4, 9, 16, 25, None, 49, 64, 81]
        range(10) | (~tryout(retries=1) | apply(f)) | deref()
        counter = 0 # line below returns [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]
        range(10) | (~tryout(retries=2) | apply(f)) | deref()

:param result: result to return if there is an exception
:param mode: if "result" (default), returns the result if there's an exception. If "str" then
    returns the exception's string. If "input" then returns the original input. If "traceback"
    then returns the exception's traceback
:param retries: how many time to retry before giving up?"""                      # tryout
        super().__init__(capture=True); self.result = result; self.inverted = False; self.retries = retries; self.mode = mode # tryout
    def __ror__(self, it):                                                       # tryout
        retries = self.retries; result = self.result; mode = self.mode           # tryout
        if len(self.capturedClis) == 0: raise Exception("tryout() currently does not wrap around any other cli. You may need to change `data | tryout() | cli1() | cli2()` into `data | (tryout() | cli1() | cli2())`") # tryout
        if not self.inverted: # single mode                                      # tryout
            while True:                                                          # tryout
                try: return it | self.capturedSerial                             # tryout
                except Exception as e:                                           # tryout
                    if retries <= 0: return str(e) if mode == "str" else (it if mode == "input" else (traceback.format_exc() if mode == "traceback" else result)) # tryout
                    retries -= 1                                                 # tryout
        else: # array mode                                                       # tryout
            def gen(it):                                                         # tryout
                patience = retries; savedInputs = k1lib.Wrapper(deque()); ogInp = None # tryout
                def interceptIt(it):                                             # tryout
                    for e in it: savedInputs().append(e); yield e                # tryout
                it = iter(it); ogIt = it; it = interceptIt(it); outIt = it | self.capturedSerial # tryout
                while True:                                                      # tryout
                    try: e = next(outIt); yield e; savedInputs().popleft(); patience = retries # tryout
                    except StopIteration: break                                  # tryout
                    except Exception as e:                                       # tryout
                        if patience <= 0: ogInp = savedInputs().popleft(); patience = retries # ran out of patience, so gonna just return the canned result instead # tryout
                        else: patience -= 1                                      # tryout
                        # restart the loop                                       # tryout
                        it = interceptIt([list(savedInputs()), ogIt] | cli.joinStreams()) # tryout
                        savedInputs.value = deque(); outIt = it | self.capturedSerial # tryout
                        if patience == retries:                                  # tryout
                            yield str(e) if mode == "str" else (ogInp if mode == "input" else (traceback.format_exc() if mode == "traceback" else result)) # just resetted # tryout
            return gen(it)                                                       # tryout
    def __invert__(self): self.inverted = not self.inverted; return self         # tryout
def resume(fn):                                                                  # resume
    """Resumes a long-running operation. I usually have code that
looks like this::

    def f(x): pass # long running, expensive calculation
    ls(".") | applyMp(f) | apply(dill.dumps) | file("somefile.pth") # executing
    cat.pickle("somefile.pth") | aS(list) # getting all of the saved objects

This will read all the files in the current directory, transforms
them using the long-running, expensive function, potentially doing
it in multiple processes. Then the results are serialized (turns into
bytes) and it will be appended to an output file.

What's frustrating is that I do stupid things all the time, so the
process usually gets interrupted. But I don't want to redo the
existing work, so that's where this cli comes into play. Now it looks
like this instead::

    ls(".") | resume("somefile.pth") | applyMp(f) | apply(dill.dumps) >> file("somefile.pth")

Note that we're inserting a resume() AND changed the file write mode
to append, so that the file doesn't get overriden. Internally, this is
just a shorthand for ``~head(fn | (tryout(0) | aS(cat.pickle) | shape(0)))``

:param fn: file name"""                                                          # resume
    return ~cli.head(fn | (cli.tryout(0) | cli.aS(cli.cat.pickle) | cli.shape(0))) # resume
class trigger(BaseCli):                                                          # trigger
    def __init__(self, col=None, before=False): # yields whenever there's a delta # trigger
        """Yields elements whenever the requested column changes in value.
Example::

    # returns [1, 2, 3, 4, 2]
    [1, 1, 1, 2, 2, 3, 4, 4, 2, 2] | trigger() | cli.deref()

    data = [[1, 0], [2, 1], [2, 2], [1, 3], [1, 4]]
    # returns [[1, 0], [2, 1], [1, 3]]
    data | trigger(0) | deref()
    # returns [[1, 0], [2, 2], [1, 4]], note how on a trigger boundary, this returns the previous ("before") value, not the next value
    data | trigger(0, True) | deref()

    data = [[1, 0], [2, 2], [2, 2], [1, 3], [1, 4]]
    # returns [[1, 0], [2, 2], [1, 3], [1, 4]]
    data | trigger([0, 1]) | deref()

This takes inspiration from electrical engineering, where a capacitor is
essentially a "trigger detector". If the input voltage changes too rapidly,
the capacitor will make the output voltage spike up or down dramatically,
as if it's taking the derivative of the input signal. Slightly related is
the Schmitt trigger.

Anyway, this cli will track whether a column changes in value, and if it
does, yields the change, else ignore it. This is useful when processing
time series data, to discard adjacent messages that are effectively identical
and carries no extra information.

The parameter .col can be one of these 3 types:

- None (default): This considers the whole element/row for value change signal
- int: This considers a single column for value change signal
- list[int]: This considers a tuple of the selected columns for value change signal

Related, but not precisely the same operation include :class:`~k1lib.cli.filt.unique`
and ``aS(set)``

:param col: column to trigger
:param before: whether to take the previous or the next value when it
    changes. Defaulted to next value (False)"""                                  # trigger
        self.col = col; self.before = before                                     # trigger
    def __ror__(self, it):                                                       # trigger
        empty = object() # random sentinel                                       # trigger
        col = self.col; before = self.before; lastValue = empty; lastRow = empty # trigger
        if col is not None:                                                      # trigger
            if isinstance(col, int): # single column                             # trigger
                if before:                                                       # trigger
                    for row in it:                                               # trigger
                        row = list(row); value = row[col]                        # trigger
                        if value != lastValue:                                   # trigger
                            if lastRow is not empty: yield lastRow               # trigger
                        lastRow = row; lastValue = value                         # trigger
                    yield row                                                    # trigger
                else:                                                            # trigger
                    for row in it:                                               # trigger
                        row = list(row); value = row[col]                        # trigger
                        if value != lastValue: yield row; lastRow = row; lastValue = value # trigger
            else: # list[int], multiple columns considered 1                     # trigger
                if before:                                                       # trigger
                    for row in it:                                               # trigger
                        row = list(row); value = tuple(row[i] for i in col)      # trigger
                        if value != lastValue:                                   # trigger
                            if lastRow is not empty: yield lastRow               # trigger
                        lastRow = row; lastValue = value                         # trigger
                    yield row                                                    # trigger
                else:                                                            # trigger
                    for row in it:                                               # trigger
                        row = list(row); value = tuple(row[i] for i in col)      # trigger
                        if value != lastValue: yield row; lastRow = row; lastValue = value # trigger
        else: # tales the whole thing in                                         # trigger
            if before:                                                           # trigger
                for value in it:                                                 # trigger
                    if value != lastValue:                                       # trigger
                        if lastValue is not empty: yield lastValue               # trigger
                    lastValue = value                                            # trigger
                yield value                                                      # trigger
            else:                                                                # trigger
                for value in it:                                                 # trigger
                    if value != lastValue: yield value; lastValue = value        # trigger
class filtStd(BaseCli):                                                          # filtStd
    def __init__(self, col:int=None, std:float=2, N:int=1):                      # filtStd
        """Filters out values that is outside the specified standard deviation.
Example::

    data = [*np.random.randn(100), *np.random.randn(10)*10] | randomize(None) | deref()
    data | filtStd(std=2)   | shape(0) # likely returns around 104
    data | filtStd(std=0.1) | shape(0) # likely returns around 22

    # column mode
    data | apply(lambda x: ["a", x]) | filtStd(1, std=2) | shape(0) # likely returns around 104

    # inverse mode
    data | ~filtStd(std=0.1) | shape(0) # likely returns around 88

:param col: column to extract the value out of
:param std: how many standard deviations above and below to accept the values
:param N: how many times to do this operation. ``filtStd(std=1, N=2)`` is equivalent to ``filtStd(std=1) | filtStd(std=1)``""" # filtStd
        self.col = col; self.std = std; self.N = N; self.inverted = False        # filtStd
    def __invert__(self): res = filtStd(self.col, self.std); res.inverted = not self.inverted; return res # filtStd
    def __ror__(self, it):                                                       # filtStd
        if self.N != 1:                                                          # filtStd
            f = filtStd(self.col, self.std, 1); f.inverted = self.inverted       # filtStd
            for i in range(self.N): it = it | f                                  # filtStd
            return it                                                            # filtStd
        col = self.col; inv = self.inverted; fStd = self.std # filter std        # filtStd
        if col is None:                                                          # filtStd
            if isinstance(it, k1lib.settings.cli.arrayTypes):                    # filtStd
                mean = it.mean(); std = it.std()                                 # filtStd
                minV = mean - std*fStd; maxV = mean + std*fStd                   # filtStd
                if not inv: return it[(it >= minV) * (it <= maxV)]               # filtStd
                else:       return it[(it <  minV) + (it >  maxV)]               # filtStd
            it = list(it); mean = it | cli.toMean(); std = it | cli.toStd()      # filtStd
            minV = mean - std*fStd; maxV = mean + std*fStd                       # filtStd
            if not inv: return [v for v in it if minV <= v <= maxV]              # filtStd
            else:       return [v for v in it if v < minV or v > maxV]           # filtStd
        else:                                                                    # filtStd
            if isinstance(it, k1lib.settings.cli.arrayTypes):                    # filtStd
                mean = it[:,col].mean(); std = it[:,col].std()                   # filtStd
                minV = mean - std*fStd; maxV = mean + std*fStd                   # filtStd
                x = it[:,col] | cli.toMean().all()                               # filtStd
                if not inv: return it[(minV <= x) * (x <= maxV)]                 # filtStd
                else:       return it[(minV >  x) + (x > maxV)]                  # filtStd
            row, it = it | cli.peek()                                            # filtStd
            if it == []: return []                                               # filtStd
            try: # sliceable? This optimization cause the input might have sliceable rows already, so don't turn them into lists, to gain perf # filtStd
                if len(row): it[0]                                               # filtStd
                it = list(it)                                                    # filtStd
            except: it = [list(row) for row in it]                               # filtStd
            x = [row[col] for row in it]                                         # filtStd
            mean = x | cli.toMean(); std = x | cli.toStd()                       # filtStd
            minV = mean - std*fStd; maxV = mean + std*fStd                       # filtStd
            if not inv: return [row for row in it if minV <= row[col] <= maxV]   # filtStd
            else:       return [row for row in it if minV > row[col] or row[col] > maxV] # filtStd



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""All tools related to GenBank file format. Expected to use behind the
"gb" module name, like this::

    from k1lib.imports import *
    cat("abc.gb") | gb.feats()
"""
from k1lib import cli
from typing import Any, Union, List
__all__ = ["feats", "origin"]
class feats(cli.BaseCli):                                                        # feats
    """Fetches features, each on a separate stream.
Example::

    cat("a.gb") | gb.feats()

Output example::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"'],
     ['     gene            11874..14409',
      '                     /gene="DDX11L1"',
      '                     /note="DEAD/H-box helicase 11 like 1 (pseudogene); Derived',
      '                     by automated computational analysis using gene prediction',
      '                     method: BestRefSeq."',
      '                     /pseudo',
      '                     /db_xref="GeneID:100287102"',
      '                     /db_xref="HGNC:HGNC:37102"']]
"""                                                                              # feats
    def __ror__(self, it):                                                       # feats
        it = it | cli.grep("FEATURES", 0, 1e9).till("ORIGIN") | cli.rows()[1:-1] # feats
        cache = []                                                               # feats
        for line in it:                                                          # feats
            if line[4:9] != "     ": # new section detected                      # feats
                if len(cache) > 0: yield cache                                   # feats
                cache = []                                                       # feats
            cache.append(line)                                                   # feats
        if len(cache) > 0: yield iter(cache)                                     # feats
    @staticmethod                                                                # feats
    def filt(*terms:str) -> cli.BaseCli:                                         # feats
        """Filters for specific terms in all the features texts. If there
are multiple terms, then filters for first term, then second, then third,
so the term's order might matter to you. Example::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"'],
     ['     gene            11874..14409',
      '                     /gene="DDX11L1"',
      '                     /note="DEAD/H-box helicase 11 like 1 (pseudogene); Derived',
      '                     by automated computational analysis using gene prediction',
      '                     method: BestRefSeq."',
      '                     /pseudo',
      '                     /db_xref="GeneID:100287102"',
      '                     /db_xref="HGNC:HGNC:37102"']] | gb.feats.filt("mol_type")

Output::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"']]
"""                                                                              # feats
        if len(terms) == 0: return cli.iden()                                    # feats
        if len(terms) > 1: return cli.deref() | cli.init.serial(*(feats.filt(term) for term in terms)) # feats
        return cli.toList().all() | cli.filt(lambda F: F | cli.grep(terms[0]) | cli.shape(0) > 0) # feats
    @staticmethod                                                                # feats
    def root() -> cli.BaseCli:                                                   # feats
        """Gets root (top most unnamed tag) of a feature.
Example::

     ['     misc_RNA        complement(join(14362..14829,14970..15038,15796..15947,',
      '                     16607..16765,16858..17055,17233..17368,17606..17742,',
      '                     17915..18061,18268..18366,24738..24891,29321..29370))',
      '                     /gene="WASH7P"',
      '                     /gene_synonym="FAM39F; WASH5P"',
      '                     /product="WASP family homolog 7, pseudogene"',
      '                     /note="Derived by automated computational analysis using',
      '                     gene prediction method: BestRefSeq."',
      '                     /pseudo',
      '                     /transcript_id="NR_024540.1"',
      '                     /db_xref="GeneID:653635"',
      '                     /db_xref="HGNC:HGNC:38034"'] | feats.root()

Output::

    ['misc_RNA',
     ['complement(join(14362..14829,14970..15038,15796..15947,',
      '16607..16765,16858..17055,17233..17368,17606..17742,',
      '17915..18061,18268..18366,24738..24891,29321..29370))']]
"""                                                                              # feats
        return cli.apply(lambda x: x.strip()) | cli.breakIf(lambda x: x.startswith("/")) | ~cli.aS(lambda a, *b: [a.split(" ")[0], [" ".join(a.split(" ")[1:]).strip(), *b]]) # feats
    @staticmethod                                                                # feats
    def tags(*tags:List[str]) -> cli.BaseCli:                                    # feats
        """Grabs a list of tags.
Example::

    s = ['     misc_RNA        complement(join(14362..14829,14970..15038,15796..15947,',
         '                     16607..16765,16858..17055,17233..17368,17606..17742,',
         '                     17915..18061,18268..18366,24738..24891,29321..29370))',
         '                     /gene="WASH7P"',
         '                     /gene_synonym="FAM39F; WASH5P"',
         '                     /product="WASP family homolog 7, pseudogene"',
         '                     /note="Derived by automated computational analysis using',
         '                     gene prediction method: BestRefSeq."',
         '                     /pseudo',
         '                     /transcript_id="NR_024540.1"',
         '                     /db_xref="GeneID:653635"',
         '                     /db_xref="HGNC:HGNC:38034"']
    s | feats.tags()

Output::

    [['gene', 'WASH7P'],
     ['gene_synonym', 'FAM39F; WASH5P'],
     ['product', 'WASP family homolog 7, pseudogene'],
     ['note',
      'Derived by automated computational analysis using gene prediction method: BestRefSeq.'],
     ['pseudo', ''],
     ['transcript_id', 'NR_024540.1'],
     ['db_xref', 'GeneID:653635'],
     ['db_xref', 'HGNC:HGNC:38034']]

With filters::

    # returns [['gene', 'WASH7P'], ['db_xref', 'HGNC:HGNC:38034'], ['organism', '']]
    s | feats.tags("gene", "db_xref", "organism")
"""                                                                              # feats
        f = cli.op().strip().all() | cli.grep("^/", sep=True).till() | cli.join(" ").all()\
            | (cli.op().split("=") | ~cli.aS(lambda a, *b: [a[1:], b[0][1:-1] if len(b) > 0 else ""])).all() # feats
        if len(tags) == 0: return f | cli.deref()                                # feats
        def g(it):                                                               # feats
            d = it | f | cli.toDict()                                            # feats
            return [[tag, d[tag] if tag in d else ""] for tag in tags]           # feats
        return cli.aS(g)                                                         # feats
class origin(cli.BaseCli):                                                       # origin
    """Return the origin section of the genbank file.
Example::

    # returns single fasta string
    cat("a.gb") | gb.origin()
"""                                                                              # origin
    def __ror__(self, it):                                                       # origin
        return it | cli.grep("ORIGIN", 0, 1e9) | ~cli.head(1) | cli.op().strip().all()\
        | cli.op().split(" ").all() | cli.cut()[1:] | cli.join("").all()\
        | cli.op().replace("/", "").all() | cli.join("")                         # origin
